"""This package implements multistart generators."""
from __future__ import annotations

from bqskit.ir.opt.multistartgens.diagonal import DiagonalStartGenerator
from bqskit.ir.opt.multistartgens.random import RandomStartGenerator

__all__ = ['DiagonalStartGenerator', 'RandomStartGenerator']



"""This module implements the ForEachBlockPass class."""
from __future__ import annotations

import functools
import logging
from typing import Callable

from bqskit.compiler.basepass import _sub_do_work
from bqskit.compiler.basepass import BasePass
from bqskit.compiler.machine import MachineModel
from bqskit.compiler.passdata import PassData
from bqskit.compiler.workflow import Workflow
from bqskit.compiler.workflow import WorkflowLike
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.circuitgate import CircuitGate
from bqskit.ir.gates.constant.unitary import ConstantUnitaryGate
from bqskit.ir.gates.parameterized.pauli import PauliGate
from bqskit.ir.gates.parameterized.unitary import VariableUnitaryGate
from bqskit.ir.location import CircuitLocation
from bqskit.ir.operation import Operation
from bqskit.ir.point import CircuitPoint
from bqskit.runtime import get_runtime

_logger = logging.getLogger(__name__)


class ForEachBlockPass(BasePass):
    """
    A pass that executes other passes on each block in the circuit.

    This is a control pass that executes a workflow on every block in the
    circuit. This will be done in parallel.
    """

    key = 'ForEachBlockPass_data'
    """The key in data, where block data will be put."""

    pass_down_key_prefix = 'ForEachBlockPass_pass_down_'
    """If a key exists in the pass data with this prefix, pass it to blocks."""

    pass_down_block_specific_key_prefix = (
        'ForEachBlockPass_specific_pass_down_'
    )
    """
    Data specific to the processing of individual blocks in a partitioned
    circuit can be injected into the `PassData` in `run` by using this prefix.

    The expected type of the associated value is `dict[int, Any]`, where
    integer (sub-)keys correspond to block numbers in a partitioned quantum
    circuit.

    Pseudocode example for seed circuits:
        seeds = {block_id: [seed_circuit_a, seed_circuit_b, ...], ...}
        key = self.pass_down_block_specific_key_prefix + 'seed_circuits'
        seed_updater = UpdateDataPass(key, seeds)
        workflow = Workflow([..., seed_updater, ForEachBlockPass(...), ...])
    """

    def __init__(
        self,
        loop_body: WorkflowLike,
        calculate_error_bound: bool = False,
        collection_filter: Callable[[Operation], bool] | None = None,
        replace_filter: ReplaceFilterFn | str = 'always',
        batch_size: int | None = None,
    ) -> None:
        """
        Construct a ForEachBlockPass.

        Args:
            loop_body (WorkflowLike): The workflow to execute on every block.

            calculate_error_bound (bool): If set to true, will calculate
                errors on blocks after running `loop_body` on them and
                use these block errors to calculate an upper bound on the
                full circuit error. (Default: False)

            collection_filter (Callable[[Operation], bool] | None):
                A predicate that determines which operations should have
                `loop_body` called on them. Called with each operation
                in the circuit. If this returns true, that operation will
                be formed into an individual circuit and passed through
                `loop_body`. Defaults to all CircuitGates,
                ConstantUnitaryGates, and VariableUnitaryGates.
                #TODO: address importability

            replace_filter (ReplaceFilterFn | str | None):
                A predicate that determines if the resulting circuit, after
                calling `loop_body` on a block, should replace the original
                operation. Called with the circuit output from `loop_body`
                and the original operation. If this returns true, the
                operation will be replaced with the new circuit.
                Defaults to always replace. If none is passed, will
                generate a replace filter always replaces. If a string is
                passed, will generate a replace filter corresponding to
                the string. The string should either be 'always', 'less-than',
                'less-than-multi', 'less-than-many', 'less-than-respecting',
                'less-than-respecting-multi', or 'less-than-respecting-many'.
                    - 'always' will always replace
                    - 'less-than' will replace if the new circuit has fewer
                        gates than the old circuit.
                    - 'less-than-multi' will replace if the new circuit has
                        fewer multi-qudit gates than the old circuit.
                    - 'less-than-many' will replace if the new circuit has
                        fewer many-qudit gates than the old circuit.
                    - 'less-than-respecting' will replace if the new circuit
                        has fewer gates than the old circuit or the old
                        doesn't respect the model (ignoring single-qudit
                        gate sets).
                    - 'less-than-respecting-multi' will replace if the new
                        circuit has fewer multi-qudit gates than the old
                        circuit or the old doesn't respect the model
                        (ignoring single-qudit gate sets).
                    - 'less-than-respecting-many' will replace if the new
                        circuit has fewer many-qudit gates than the old
                        circuit or the old doesn't respect the model
                        (ignoring single-qudit gate sets).
                    - 'less-than-respecting-fully' will replace if the new
                        circuit has fewer gates than the old circuit or
                        the old doesn't respect the model.
                    - 'less-than-respecting-fully-multi' will replace if
                        the new circuit has fewer multi-qudit gates than
                        the old circuit or the old doesn't respect the model.
                    - 'less-than-respecting-fully-many' will replace if
                        the new circuit has fewer many-qudit gates than
                        the old circuit or the old doesn't respect the model.
                Defaults to 'always'.  #TODO: address importability

            batch_size (int): (Deprecated).
        """
        if batch_size is not None:
            import warnings
            warnings.warn(
                'Batch size is no longer supported, this warning will'
                ' become an error in a future update.',
                DeprecationWarning,
            )

        self.calculate_error_bound = calculate_error_bound
        self.collection_filter = collection_filter or default_collection_filter
        self.replace_filter = replace_filter or default_replace_filter
        self.workflow = Workflow(loop_body)

        if not callable(self.collection_filter):
            raise TypeError(
                'Expected callable method that maps Operations to booleans for'
                f' collection_filter, got {type(self.collection_filter)}.',
            )

        if not isinstance(self.replace_filter, str):
            if not callable(self.replace_filter):
                raise TypeError(
                    'Expected either string representing a valid replacement'
                    ' filter or callable method that maps Circuit and'
                    ' Operations to bools for replace_filter'
                    f' , got {type(self.replace_filter)}.',
                )

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        # Get the callable replacement filter
        if isinstance(self.replace_filter, str):
            method = self.replace_filter
            replace_filter = gen_replace_filter(method, data.model)
        else:
            replace_filter = self.replace_filter

        # Make room in data for block data
        if self.key not in data:
            data[self.key] = []

        # Collect blocks
        blocks: list[tuple[int, Operation]] = []
        for cycle, op in circuit.operations_with_cycles():
            if self.collection_filter(op):
                blocks.append((cycle, op))

        # No blocks, no work
        if len(blocks) == 0:
            data[self.key].append([])
            return

        # Get the machine model
        model = data.model
        coupling_graph = data.connectivity

        # Preprocess blocks
        subcircuits: list[Circuit] = []
        block_datas: list[PassData] = []
        for i, (cycle, op) in enumerate(blocks):

            # Form Subcircuit
            if isinstance(op.gate, CircuitGate):
                subcircuit = op.gate._circuit.copy()
                subcircuit.set_params(op.params)
            else:
                subcircuit = Circuit.from_operation(op)

            # Form Submodel
            subradixes = [circuit.radixes[q] for q in op.location]
            subnumbering = {op.location[i]: i for i in range(len(op.location))}
            submodel = MachineModel(
                len(op.location),
                coupling_graph.get_subgraph(op.location, subnumbering),
                model.gate_set,
                subradixes,
            )

            # Form Subdata
            block_data: PassData = PassData(subcircuit)
            block_data['subnumbering'] = subnumbering
            block_data['model'] = submodel
            block_data['point'] = CircuitPoint(cycle, op.location[0])
            block_data['calculate_error_bound'] = self.calculate_error_bound
            for key in data:
                if key.startswith(self.pass_down_key_prefix):
                    block_data[key] = data[key]
                elif key.startswith(
                    self.pass_down_block_specific_key_prefix,
                ) and i in data[key]:
                    block_data[key] = data[key][i]
            block_data.seed = data.seed

            subcircuits.append(subcircuit)
            block_datas.append(block_data)

        # Do the work
        results = await get_runtime().map(
            _sub_do_work,
            [self.workflow] * len(subcircuits),
            subcircuits,
            block_datas,
        )

        # Unpack results
        completed_subcircuits, completed_block_datas = zip(*results)

        # Postprocess blocks
        points: list[CircuitPoint] = []
        ops: list[Operation] = []
        error_sum = 0.0
        for i, (cycle, op) in enumerate(blocks):
            subcircuit = completed_subcircuits[i]
            block_data = completed_block_datas[i]

            # Mark Blocks to be Replaced
            if replace_filter(subcircuit, op):
                _logger.debug(f'Replacing block {i}.')
                points.append(CircuitPoint(cycle, op.location[0]))
                ops.append(
                    Operation(
                        CircuitGate(subcircuit, True),
                        op.location,
                        subcircuit.params,
                    ),
                )
                block_data['replaced'] = True

                # Calculate Error
                error_sum += block_data.error
            else:
                block_data['replaced'] = False

        # Replace blocks
        circuit.batch_replace(points, ops)

        # Record block data into pass data
        data[self.key].append(completed_block_datas)

        # Record error
        data.update_error_mul(error_sum)
        if self.calculate_error_bound:
            _logger.debug(f'New circuit error is {data.error}.')


def default_collection_filter(op: Operation) -> bool:
    return isinstance(
        op.gate, (
            CircuitGate,
            ConstantUnitaryGate,
            VariableUnitaryGate,
            PauliGate,
        ),
    )


def default_replace_filter(circuit: Circuit, op: Operation) -> bool:
    """Always replace."""
    # legacy name and style for backwards compatibility
    return True


def _less_than(new: Circuit, old: Operation) -> bool:
    """Return true if the new circuit has fewer gates."""
    if isinstance(old.gate, CircuitGate):
        return new.num_operations < old.gate._circuit.num_operations

    return True  # TODO: Re-evaluate always true when old is not a circuit


def _less_than_multi(new: Circuit, old: Operation) -> bool:
    """Return true if the new circuit has fewer multi-qudit gates."""
    if isinstance(old.gate, CircuitGate):
        org = old.gate._circuit
        omq = sum([c for g, c in org.gate_counts.items() if g.num_qudits > 1])
        osq = sum([c for g, c in org.gate_counts.items() if g.num_qudits == 1])
        nmq = sum([c for g, c in new.gate_counts.items() if g.num_qudits > 1])
        nsq = sum([c for g, c in new.gate_counts.items() if g.num_qudits == 1])
        return (nmq, nsq) < (omq, osq)

    return True


def _less_than_many(new: Circuit, old: Operation) -> bool:
    """Return true if the new circuit has fewer many-qudit gates."""
    if isinstance(old.gate, CircuitGate):
        org = old.gate._circuit
        omq = sum([c for g, c in org.gate_counts.items() if g.num_qudits > 2])
        otq = sum([c for g, c in org.gate_counts.items() if g.num_qudits == 2])
        osq = sum([c for g, c in org.gate_counts.items() if g.num_qudits == 1])
        nmq = sum([c for g, c in new.gate_counts.items() if g.num_qudits > 2])
        ntq = sum([c for g, c in new.gate_counts.items() if g.num_qudits == 2])
        nsq = sum([c for g, c in new.gate_counts.items() if g.num_qudits == 1])
        return (nmq, ntq, nsq) < (omq, otq, osq)

    return True


def _is_respecting(
    circuit: Circuit,
    location: CircuitLocation,
    model: MachineModel,
    fully: bool = False,
) -> bool:
    """
    Return true if the `circuit` respects the `model` at `location`.

    Args:
        circuit (Circuit): The circuit to check.

        location (CircuitLocation): The location to check.

        model (MachineModel): The machine model to check against.

        fully (bool): If set to true, will check if the circuit respects
            the model fully. If set to false, will ignore single-qudit
            gate sets. (Default: False)

    Returns:
        True if the circuit respects the model at the location. This implies
        that the circuit can be run on the machine at the location.
    """
    org_mq_gates = circuit.gate_set.multi_qudit_gates
    org_sq_gates = circuit.gate_set.single_qudit_gates

    if any(g not in model.gate_set for g in org_mq_gates):
        return False

    if fully and any(g not in model.gate_set for g in org_sq_gates):
        return False

    if any(
        (location[e[0]], location[e[1]]) not in model.coupling_graph
        for e in circuit.coupling_graph
    ):
        return False

    return True


def _less_than_fn_respecting(
    new: Circuit,
    old: Operation,
    model: MachineModel,
    fn: ReplaceFilterFn,
) -> bool:
    """Return true if the new circuit has fewer gates or the old doesn't respect
    the model."""
    if isinstance(old.gate, CircuitGate):
        if not _is_respecting(old.gate._circuit, old.location, model):
            if not _is_respecting(new, old.location, model):
                _logger.debug("New block doesn't respect model.")
            return True

        if not _is_respecting(new, old.location, model):
            _logger.debug("New block doesn't respect model.")
            return False

    return fn(new, old)


def _less_than_fn_respecting_fully(
    new: Circuit,
    old: Operation,
    model: MachineModel,
    fn: ReplaceFilterFn,
) -> bool:
    """Return true if the new circuit has fewer gates or the old doesn't respect
    the model."""
    if isinstance(old.gate, CircuitGate):
        if not _is_respecting(old.gate._circuit, old.location, model, True):
            if not _is_respecting(new, old.location, model, True):
                _logger.debug("New block doesn't respect model.")
            return True

        if not _is_respecting(new, old.location, model, True):
            _logger.debug("New block doesn't respect model.")
            return False

    return fn(new, old)


def gen_always(model: MachineModel) -> ReplaceFilterFn:
    """Generate a replace filter that always replaces."""
    # legacy name and style for backwards compatibility
    return default_replace_filter


def gen_less_than(model: MachineModel) -> ReplaceFilterFn:
    """Generate a replace filter that replaces if the new circuit has fewer
    gates."""
    return _less_than


def gen_less_than_multi(model: MachineModel) -> ReplaceFilterFn:
    """Generate a replace filter that replaces if the new circuit has fewer
    multi-qudit gates."""
    return _less_than_multi


def gen_less_than_many(model: MachineModel) -> ReplaceFilterFn:
    """Generate a replace filter that replaces if the new circuit has fewer
    many-qudit gates."""
    return _less_than_many


def gen_less_than_rspt(model: MachineModel) -> ReplaceFilterFn:
    """Generate a replace filter that replaces if the new circuit has fewer
    gates or the old doesn't respect the model."""
    return functools.partial(
        _less_than_fn_respecting,
        model=model,
        fn=_less_than,
    )


def gen_less_than_rspt_multi(model: MachineModel) -> ReplaceFilterFn:
    """Generate a replace filter that replaces if the new circuit has fewer
    multi-qudit gates or the old doesn't respect the model."""
    return functools.partial(
        _less_than_fn_respecting,
        model=model,
        fn=_less_than_multi,
    )


def gen_less_than_rspt_many(model: MachineModel) -> ReplaceFilterFn:
    """Generate a replace filter that replaces if the new circuit has fewer
    many-qudit gates or the old doesn't respect the model."""
    return functools.partial(
        _less_than_fn_respecting,
        model=model,
        fn=_less_than_many,
    )


def gen_less_than_rspt_fully(model: MachineModel) -> ReplaceFilterFn:
    """Generate a replace filter that replaces if the new circuit has fewer
    gates or the old doesn't respect the model."""
    return functools.partial(
        _less_than_fn_respecting_fully,
        model=model,
        fn=_less_than,
    )


def gen_less_than_rspt_fully_multi(model: MachineModel) -> ReplaceFilterFn:
    """Generate a replace filter that replaces if the new circuit has fewer
    multi-qudit gates or the old doesn't respect the model."""
    return functools.partial(
        _less_than_fn_respecting_fully,
        model=model,
        fn=_less_than_multi,
    )


def gen_less_than_rspt_fully_many(model: MachineModel) -> ReplaceFilterFn:
    """Generate a replace filter that replaces if the new circuit has fewer
    many-qudit gates or the old doesn't respect the model."""
    return functools.partial(
        _less_than_fn_respecting_fully,
        model=model,
        fn=_less_than_many,
    )


def gen_replace_filter(method: str, model: MachineModel) -> ReplaceFilterFn:
    """
    Generate a replace filter for use during the standard workflow.

    Args:
        method (str): The method to use for the replace filter. See
            :class:`ForEachBlockPass` for more information.

        model (MachineModel): The machine model to potentially respect.

    Returns:
        A replace filter function.
    """
    replace_filters = {
        'always': gen_always,
        'less-than': gen_less_than,
        'less-than-multi': gen_less_than_multi,
        'less-than-many': gen_less_than_many,
        'less-than-respecting': gen_less_than_rspt,
        'less-than-respecting-multi': gen_less_than_rspt_multi,
        'less-than-respecting-many': gen_less_than_rspt_many,
        'less-than-respecting-fully': gen_less_than_rspt_fully,
        'less-than-respecting-fully-multi': gen_less_than_rspt_fully_multi,
        'less-than-respecting-fully-many': gen_less_than_rspt_fully_many,
    }

    if method not in replace_filters:
        raise ValueError(f'Unknown replace filter method {method}.')

    return replace_filters[method](model)


ReplaceFilterFn = Callable[[Circuit, Operation], bool]


class ClearAllBlockData(BasePass):
    """Clear all block data and passed down data from the pass data."""

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        for key in list(data.keys()):
            if key.startswith(ForEachBlockPass.key):
                del data[key]
            elif key.startswith(ForEachBlockPass.pass_down_key_prefix):
                del data[key]



"""This module implements the ExhaustiveGateRemovalPass."""
from __future__ import annotations

import logging
from typing import Any
from typing import Callable

import numpy as np

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.operation import Operation
from bqskit.ir.opt.cost.functions import HilbertSchmidtResidualsGenerator
from bqskit.ir.opt.cost.generator import CostFunctionGenerator
from bqskit.ir.structure import CircuitStructure
from bqskit.runtime import get_runtime
from bqskit.utils.typing import is_real_number
_logger = logging.getLogger(__name__)


class ExhaustiveGateRemovalPass(BasePass):
    """
    The ExhaustiveGateRemovalPass class.

    Use instantiation to remove the most possible gates from the circuit.
    """

    def __init__(
        self,
        success_threshold: float = 1e-8,
        cost: CostFunctionGenerator = HilbertSchmidtResidualsGenerator(),
        instantiate_options: dict[str, Any] = {},
        collection_filter: Callable[[Operation], bool] | None = None,
        scoring_fn: Callable[[Circuit], float] | None = None,
    ) -> None:
        """
        Construct a ExhaustiveGateRemovalPass.

        Args:

            success_threshold (float): The distance threshold that
                determines successful termintation. Measured in cost
                described by the hilbert schmidt cost function.
                (Default: 1e-8)

            cost (CostFunction | None): The cost function that determines
                successful removal of a gate.
                (Default: HilbertSchmidtResidualsGenerator())

            instantiate_options (dict[str: Any]): Options passed directly
                to circuit.instantiate when instantiating circuit
                templates. (Default: {})

            collection_filter (Callable[[Operation], bool] | None):
                A predicate that determines which operations should be
                attempted to be removed. Called with each operation
                in the circuit. If this returns true, this pass will
                attempt to remove that operation. Defaults to all
                operations.

            scoring_fn (Callable[[Circuit], float]):
                A scoring function for the circuits to determine which one
                to select. Defaults to gate counts weighted by their size.
        """

        if not is_real_number(success_threshold):
            raise TypeError(
                'Expected real number for success_threshold'
                ', got %s' % type(success_threshold),
            )

        if not isinstance(cost, CostFunctionGenerator):
            raise TypeError(
                'Expected cost to be a CostFunctionGenerator, got %s'
                % type(cost),
            )

        if not isinstance(instantiate_options, dict):
            raise TypeError(
                'Expected dictionary for instantiate_options, got %s.'
                % type(instantiate_options),
            )

        self.collection_filter = collection_filter or default_collection_filter

        if not callable(self.collection_filter):
            raise TypeError(
                'Expected callable method that maps Operations to booleans for'
                ' collection_filter, got %s.' % type(self.collection_filter),
            )

        self.scoring_fn = scoring_fn or default_scoring_fn

        if not callable(self.scoring_fn):
            raise TypeError(
                'Expected callable method that maps Circuits to floats for'
                ' scoring_fn, got %s.' % type(self.scoring_fn),
            )

        self.success_threshold = success_threshold
        self.cost = cost
        self.instantiate_options: dict[str, Any] = {
            'dist_tol': self.success_threshold,
            'min_iters': 100,
        }
        self.instantiate_options.update(instantiate_options)

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        instantiate_options = self.instantiate_options.copy()
        if 'seed' not in instantiate_options:
            instantiate_options['seed'] = data.seed

        _logger.debug('Starting exhaustive gate removal.')

        target = self.get_target(circuit, data)

        # Frontier tracks circuits successfully instantiated to target
        frontier = [circuit.copy()]

        # Track best cicuit seen so far
        best_circuit = None
        best_score = -np.inf

        # Keep removing until no more successful circuits
        while len(frontier) > 0:

            # Expand each element of frontier by removing gates
            expanded_circuits = []

            # Don't repeat circuit structures
            circuits_seen = set()

            for c in frontier:
                for cycle, op in c.operations_with_cycles():
                    point = (cycle, op.location[0])
                    copy = c.copy()
                    copy.pop(point)
                    structure = CircuitStructure(copy)
                    if structure not in circuits_seen:
                        expanded_circuits.append(copy)
                        circuits_seen.add(structure)

            # Instantiate them all
            instantiated_circuits = await get_runtime().map(
                Circuit.instantiate,
                expanded_circuits,
                target=target,
                **instantiate_options,
            )

            # Process them
            next_frontier = []
            for c in instantiated_circuits:
                if self.cost(c, target) < self.success_threshold:
                    next_frontier.append(c)

                    score = self.scoring_fn(c)
                    if score > best_score:
                        best_circuit = c
                        best_score = score

            frontier = next_frontier

        # Keep best circuit if one found
        if best_circuit is not None:
            circuit.become(best_circuit)


def default_collection_filter(op: Operation) -> bool:
    return True


def default_scoring_fn(circuit: Circuit) -> float:
    """Default scoring function."""
    score = 0.0
    for op in circuit:
        score -= (op.num_qudits - 1) * 100 + 1
    return score



"""This module implements the QFASTDecompositionPass class."""
from __future__ import annotations

import copy
import logging
from typing import Any
from typing import Sequence

from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gate import Gate
from bqskit.ir.gates.composed.vlg import VariableLocationGate
from bqskit.ir.gates.parameterized.pauli import PauliGate
from bqskit.ir.location import CircuitLocation
from bqskit.ir.operation import Operation
from bqskit.ir.opt.cost import CostFunctionGenerator
from bqskit.ir.opt.cost import HilbertSchmidtResidualsGenerator
from bqskit.passes.synthesis.synthesis import SynthesisPass
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_real_number

_logger = logging.getLogger(__name__)


class QFASTDecompositionPass(SynthesisPass):
    """
    A pass performing one round of decomposition from the QFAST algorithm.

    References:
        E. Younis, K. Sen, K. Yelick and C. Iancu, "QFAST: Conflating Search
        and Numerical Optimization for Scalable Quantum Circuit Synthesis,"
        2021 IEEE International Conference on Quantum Computing and
        Engineering (QCE), 2021, pp. 232-243, doi: 10.1109/QCE52317.2021.00041.
    """

    def __init__(
        self,
        gate: Gate = PauliGate(2),
        success_threshold: float = 1e-8,
        progress_threshold: float = 5e-3,
        cost: CostFunctionGenerator = HilbertSchmidtResidualsGenerator(),
        max_depth: int | None = None,
        instantiate_options: dict[str, Any] = {},
    ) -> None:
        """
        QFASTDecompositionPass Constructor.

        Args:
            gate (Gate): The gate to decompose unitaries into. Ensure
                that the gate specified is expressive over the unitaries
                being synthesized. (Default: PauliGate(2))

            success_threshold (float): The distance threshold that
                determines successful termintation. Measured in cost
                described by the cost function. (Default: 1e-8)

            progress_threshold (float): The distance necessary to improve
                for the synthesis algorithm to complete a layer and move
                on. Lowering this will led to synthesis going deeper quicker,
                and raising it will force to algorithm to spend more time
                on each layer. Caution, changing this too much might break
                the synthesis algorithm. (Default: 5e-3)

            cost (CostFunction | None): The cost function that determines
                distance during synthesis. The goal of this synthesis pass
                is to implement circuits for the given unitaries that have
                a cost less than the `success_threshold`.
                (Default: HSDistance())

            max_depth (int): The maximum number of gates to append without
                success before termination. If left as None it will default
                to unlimited. (Default: None)

            instantiate_options (dict[str: Any]): Options passed directly
                to circuit.instantiate when instantiating circuit
                templates. (Default: {})

        Raises:
            ValueError: If max_depth is nonpositive.
        """

        if not isinstance(gate, Gate):
            raise TypeError('Expected gate to be a Gate, got %s' % type(gate))

        if not is_real_number(success_threshold):
            raise TypeError(
                'Expected real number for success_threshold'
                ', got %s' % type(success_threshold),
            )

        if not is_real_number(progress_threshold):
            raise TypeError(
                'Expected real number for progress_threshold'
                ', got %s' % type(progress_threshold),
            )

        if not isinstance(cost, CostFunctionGenerator):
            raise TypeError(
                'Expected cost to be a CostFunctionGenerator, got %s'
                % type(cost),
            )

        if max_depth is not None and not is_integer(max_depth):
            raise TypeError(
                'Expected max_depth to be an integer, got %s' % type(
                    max_depth,
                ),
            )

        if max_depth is not None and max_depth <= 0:
            raise ValueError(
                'Expected max_depth to be positive, got %d.' % int(max_depth),
            )

        self.gate = gate
        self.success_threshold = success_threshold
        self.progress_threshold = progress_threshold
        self.cost = cost
        self.max_depth = max_depth
        self.instantiate_options: dict[str, Any] = {
            'cost_fn_gen': self.cost,
            'method': 'minimization',
        }
        self.instantiate_options.update(instantiate_options)

    async def synthesize(
        self,
        utry: UnitaryMatrix | StateVector | StateSystem,
        data: PassData,
    ) -> Circuit:
        """Synthesize `utry`, see :class:`SynthesisPass` for more."""
        instantiate_options = self.instantiate_options.copy()
        if 'seed' not in instantiate_options:
            instantiate_options['seed'] = data.seed

        # Skip any unitaries too small for the configured gate.
        if self.gate.num_qudits > utry.num_qudits:
            raise RuntimeError('Gate is too small for circuit')

        # Create empty circuit with same size and radixes as `utry`.
        circuit = Circuit(utry.num_qudits, utry.radixes)

        # Calculate relevant coupling_graph and create the VLG head.
        model = self.get_model(utry, data)
        locations = model.get_locations(self.gate.num_qudits)
        vlg_head = VariableLocationGate(self.gate, locations, circuit.radixes)
        circuit.append_gate(vlg_head, list(range(utry.num_qudits)))

        # Track depth and distances
        depth = 1
        last_dist = 1.0
        failed_locs: list[tuple[CircuitLocation, float]] = []

        # Main loop
        while True:

            # Instantiate circuit
            circuit.instantiate(utry, **instantiate_options)
            dist = self.cost.calc_cost(circuit, utry)
            _logger.info(f'Instantiated depth {depth} at {dist} cost.')

            if dist < self.success_threshold:
                self.finalize(circuit, utry, instantiate_options)
                _logger.info(f'Successful synthesis with {depth} layers.')
                return circuit

            # Expand or restrict head
            location = self.get_location_of_head(circuit)

            if last_dist - dist >= self.progress_threshold:
                _logger.info('Progress has been made, depth increasing.')
                last_dist = dist
                self.expand(circuit, location, locations)
                depth += 1
                failed_locs = []

            elif not self.can_restrict(circuit[-1, 0]):
                _logger.info('Progress has not been made.')
                _logger.info('Cannot restrict further, depth increasing.')
                failed_locs.append((location, dist))
                failed_locs.sort(key=lambda x: x[1])
                location, last_dist = failed_locs[0]
                self.expand(circuit, location, locations)
                depth += 1
                failed_locs = []

            else:
                _logger.info('Progress has not been made, restricting model.')
                failed_locs.append((location, dist))
                self.restrict_head(circuit, location)

    def get_location_of_head(self, circuit: Circuit) -> CircuitLocation:
        """Return the current location of the `circuit`'s VLG head."""
        head_gate: VariableLocationGate = circuit[-1, 0].gate  # type: ignore
        return CircuitLocation(head_gate.get_location(circuit[-1, 0].params))

    def finalize(
        self,
        circuit: Circuit,
        utry: UnitaryMatrix | StateVector | StateSystem,
        instantiate_options: dict[str, Any],
    ) -> None:
        """Finalize the circuit by replacing the head with self.gate."""
        # Replace Head with self.gate
        location = self.get_location_of_head(circuit)
        _logger.info(f'Final gate added at location {location}.')
        circuit.pop()
        circuit.append(Operation(self.gate, location))

        # Reinstantiate
        dist = self.cost.calc_cost(circuit, utry)
        while dist > self.success_threshold:
            circuit.instantiate(utry, **instantiate_options)
            dist = self.cost.calc_cost(circuit, utry)

        _logger.info(f'Final circuit found with cost: {dist}.')

    def expand(
        self,
        circuit: Circuit,
        location: CircuitLocation,
        locations: Sequence[CircuitLocation],
    ) -> None:
        """Expand the circuit after a successful layer."""
        _logger.info(f'Expanding by adding a gate on qubits {location}.')
        circuit.insert(-1, Operation(self.gate, location))
        self.lift_head_restrictions(circuit, locations)
        self.restrict_head(circuit, location)

    def can_restrict(self, head: Operation) -> bool:
        """Return true if the VLG head can be restricted further."""
        head_gate: VariableLocationGate = head.gate  # type: ignore
        return len(head_gate.locations) > 1

    def restrict_head(
        self,
        circuit: Circuit,
        location: CircuitLocation,
    ) -> None:
        """
        Remove `location` from the VLG Head in `circuit`.

        Args:
            circuit (Circuit): The circuit to restrict its VLG head.

            location (CircuitLocation): The location to remove from the
                VLG head.
        """
        _logger.debug(f'Removing location {location} from the VLG head.')
        head_gate: VariableLocationGate = circuit[-1, 0].gate  # type: ignore
        locations = copy.deepcopy(head_gate.locations)
        locations.remove(location)
        new_vlg = VariableLocationGate(self.gate, locations, circuit.radixes)
        new_head = Operation(new_vlg, list(range(circuit.num_qudits)))
        circuit.pop()
        circuit.append(new_head)

    def lift_head_restrictions(
        self,
        circuit: Circuit,
        locations: Sequence[CircuitLocation],
    ) -> None:
        """Set the `circuit`'s VLG head's valid locations to `locations`."""
        _logger.debug('Lifting restrictions.')
        _logger.debug(f'Reset VLG head location pool to {locations}.')
        vlg_gate = VariableLocationGate(self.gate, locations, circuit.radixes)
        new_head = Operation(vlg_gate, list(range(circuit.num_qudits)))
        circuit.pop()
        circuit.append(new_head)



from __future__ import annotations

from bqskit.ir.opt.cost.functions.cost.hilbertschmidt import (
    HilbertSchmidtCost,
)
from bqskit.ir.opt.cost.functions.cost.hilbertschmidt import (
    HilbertSchmidtCostGenerator,
)

__all__ = ['HilbertSchmidtCost', 'HilbertSchmidtCostGenerator']



"""This module implements the Workflow class."""
from __future__ import annotations

import copy
import logging
from typing import Iterable
from typing import Iterator
from typing import overload
from typing import Sequence
from typing import TYPE_CHECKING
from typing import Union

import dill

from bqskit.compiler.basepass import BasePass
from bqskit.utils.random import seed_random_sources
from bqskit.utils.typing import is_iterable

if TYPE_CHECKING:
    from bqskit.compiler.passdata import PassData
    from bqskit.ir.circuit import Circuit


_logger = logging.getLogger(__name__)


class Workflow(BasePass, Sequence[BasePass]):
    """A BQSKit workflow captures a quantum circuit compilation process."""

    def __init__(self, passes: WorkflowLike, name: str = '') -> None:
        """
        Initialize a workflow object from a sequence of passes.

        Args:
            passes (WorkflowLike): The passes to run in the workflow.

            name (str): The optional name of the workflow.

        Raises:
            ValueError: If passes is empty.
        """
        if isinstance(passes, Workflow):
            self._passes: list[BasePass] = copy.deepcopy(passes._passes)
            self._name: str = name if name else copy.deepcopy(passes._name)
            return

        if isinstance(passes, BasePass):
            passes = [passes]

        if not is_iterable(passes):
            msg = f'Expected Pass or sequence of Passes, got {type(passes)}.'
            raise TypeError(msg)

        if not all(isinstance(p, BasePass) for p in passes):
            truth_list = [isinstance(p, BasePass) for p in passes]
            wrong_type = type(list(passes)[truth_list.index(False)])
            msg = f'Expected Pass or sequence of Passes, got {wrong_type}.'
            raise TypeError(msg)

        if not isinstance(name, str):
            raise TypeError(f'Expected name to be str, got {type(name)}.')

        self._name = name
        self._passes = list(passes)

        if len(self._passes) == 0:
            raise ValueError('Expected at least one pass in workflow.')

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        for pass_obj in self._passes:
            if data.seed is not None:
                seed_random_sources(data.seed)
            _logger.debug(f'Running {pass_obj.name}')
            await pass_obj.run(circuit, data)

    def save(self, filename: str) -> None:
        import pickle
        with open(filename, 'wb') as f:
            pickle.dump(self, f)

    @staticmethod
    def load(filename: str) -> Workflow:
        import pickle
        with open(filename, 'rb') as f:
            return pickle.load(f)

    @property
    def name(self) -> str:
        """The name of the pass."""
        return self._name or self.__class__.__name__

    @staticmethod
    def is_workflow(workflow: WorkflowLike) -> bool:
        if not is_iterable(workflow):
            return isinstance(workflow, BasePass)
        return all(isinstance(p, BasePass) for p in workflow)

    def __str__(self) -> str:
        name_seq = f'Workflow: {self.name}\n\t'
        pass_strs = [
            f'{i}. Workflow: {p.name if isinstance(p, Workflow) else p}'
            for i, p in enumerate(self._passes)
        ]
        return name_seq + '\n\t'.join(pass_strs)

    def __add__(self, other: WorkflowLike) -> Workflow:
        return Workflow(self._passes + Workflow(other)._passes)

    def __radd__(self, other: WorkflowLike) -> Workflow:
        return Workflow(Workflow(other)._passes + self._passes)

    def __len__(self) -> int:
        return self._passes.__len__()

    def __iter__(self) -> Iterator[BasePass]:
        return self._passes.__iter__()

    @overload
    def __getitem__(self, _key: int, /) -> BasePass:
        ...

    @overload
    def __getitem__(self, _key: slice, /) -> list[BasePass]:
        ...

    def __getitem__(self, _key: int | slice) -> BasePass | list[BasePass]:
        return self._passes.__getitem__(_key)

    def __getstate__(self) -> bytes:
        return dill.dumps(self.__dict__, recurse=True)

    def __setstate__(self, state: bytes) -> None:
        self.__dict__.update(dill.loads(state))


WorkflowLike = Union[Workflow, Iterable[BasePass], BasePass]



from __future__ import annotations

from bqskit.ir.opt.cost.functions.residuals.hilbertschmidt import (
    HilbertSchmidtResiduals,
)
from bqskit.ir.opt.cost.functions.residuals.hilbertschmidt import (
    HilbertSchmidtResidualsGenerator,
)

__all__ = ['HilbertSchmidtResiduals', 'HilbertSchmidtResidualsGenerator']



"""This module implements the TreeScanningGateRemovalPass."""
from __future__ import annotations

import logging
from typing import Any
from typing import Callable

from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.operation import Operation
from bqskit.ir.opt.cost.functions import HilbertSchmidtResidualsGenerator
from bqskit.ir.opt.cost.generator import CostFunctionGenerator
from bqskit.passes.processing.scan import ScanningGateRemovalPass
from bqskit.runtime import get_runtime
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_real_number

_logger = logging.getLogger(__name__)


class TreeScanningGateRemovalPass(ScanningGateRemovalPass):
    """
    The TreeScanningGateRemovalPass class.

    Starting from one side of the circuit, run the following:

    Split the circuit operations into chunks of size `tree_depth`
    At every iteration:
    a. Look at the next chunk of operations
    b. Generate 2 ^ `tree_depth` circuits. Each circuit corresponds to every
    combination of whether or not to include one of the operations in the chunk.
    c. Instantiate in parallel all 2^`tree_depth` circuits
    d. Choose the circuit that has the least number of operations and move
    on to the next chunk of operations.

    This optimization is less greedy than the current
    :class:`~bqskit.passes.processing.ScanningGateRemovalPass` removal,
    which leads to much better quality circuits than ScanningGate.
    In very rare occasions, ScanningGate may be able to outperform
    TreeScan (since it is still greedy), but in general we can expect
    TreeScan to almost always outperform ScanningGate.
    """

    def __init__(
        self,
        start_from_left: bool = True,
        success_threshold: float = 1e-8,
        cost: CostFunctionGenerator = HilbertSchmidtResidualsGenerator(),
        instantiate_options: dict[str, Any] = {},
        tree_depth: int = 1,
        collection_filter: Callable[[Operation], bool] | None = None,
    ) -> None:
        """
        Construct a TreeScanningGateRemovalPass.

        Args:
            start_from_left (bool): Determines where the scan starts
                attempting to remove gates from. If True, scan goes left
                to right, otherwise right to left. (Default: True)

            success_threshold (float): The distance threshold that
                determines successful termintation. Measured in cost
                described by the hilbert schmidt cost function.
                (Default: 1e-8)

            cost (CostFunction | None): The cost function that determines
                successful removal of a gate.
                (Default: HilbertSchmidtResidualsGenerator())

            instantiate_options (dict[str: Any]): Options passed directly
                to circuit.instantiate when instantiating circuit
                templates. (Default: {})

            tree_depth (int): The depth of the tree of potential
                solutions to instantiate. Note that 2^(tree_depth) - 1
                circuits will be instantiated in parallel. Note that the default
                behavior will be equivalent to normal ScanningGateRemoval
                (Default: 1)

            collection_filter (Callable[[Operation], bool] | None):
                A predicate that determines which operations should be
                attempted to be removed. Called with each operation
                in the circuit. If this returns true, this pass will
                attempt to remove that operation. Defaults to all
                operations.
        """

        if not is_real_number(success_threshold):
            raise TypeError(
                'Expected real number for success_threshold'
                ', got %s' % type(success_threshold),
            )

        if not isinstance(cost, CostFunctionGenerator):
            raise TypeError(
                'Expected cost to be a CostFunctionGenerator, got %s'
                % type(cost),
            )

        if not isinstance(instantiate_options, dict):
            raise TypeError(
                'Expected dictionary for instantiate_options, got %s.'
                % type(instantiate_options),
            )

        self.collection_filter = collection_filter or default_collection_filter

        if not callable(self.collection_filter):
            raise TypeError(
                'Expected callable method that maps Operations to booleans for'
                ' collection_filter, got %s.' % type(self.collection_filter),
            )

        if not is_integer(tree_depth):
            raise TypeError(
                'Expected Integer type for tree_depth, got %s.'
                % type(instantiate_options),
            )

        self.tree_depth = tree_depth
        self.start_from_left = start_from_left
        self.success_threshold = success_threshold
        self.cost = cost
        self.instantiate_options: dict[str, Any] = {
            'dist_tol': self.success_threshold,
            'min_iters': 10,
            'cost_fn_gen': self.cost,
        }
        self.instantiate_options.update(instantiate_options)

    @staticmethod
    def get_tree_circs(
        orig_num_cycles: int,
        circuit_copy: Circuit,
        cycle_and_ops: list[tuple[int, Operation]],
    ) -> list[Circuit]:
        """
        Generate all circuits to be instantiated in the tree scan.

        Args:
            orig_num_cycles (int): The original number of cycles
            in the circuit. This allows us to keep track of the shift
            caused by previous deletions.

            circuit_copy (Circuit): Current state of the circuit.

            cycle_and_ops: list[(int, Operation)]: The next chunk
            of operations to be considered for deletion.

        Returns:
            list[Circuit]: A list of 2^(`tree_depth`) - 1 circuits
            that remove up to `tree_depth` operations. The circuits
            are sorted by the number of operations removed.
        """
        all_circs = [circuit_copy.copy()]
        for cycle, op in cycle_and_ops:
            new_circs = []
            for circ in all_circs:
                idx_shift = orig_num_cycles - circ.num_cycles
                new_cycle = cycle - idx_shift
                work_copy = circ.copy()
                work_copy.pop((new_cycle, op.location[0]))
                new_circs.append(work_copy)
                new_circs.append(circ)

            all_circs = new_circs

        all_circs = sorted(all_circs, key=lambda x: x.num_operations)
        # Remove circuit with no gates deleted
        return all_circs[:-1]

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        instantiate_options = self.instantiate_options.copy()
        if 'seed' not in instantiate_options:
            instantiate_options['seed'] = data.seed

        start = 'left' if self.start_from_left else 'right'
        _logger.debug(f'Starting tree scanning gate removal on the {start}.')

        target = self.get_target(circuit, data)

        circuit_copy = circuit.copy()
        reverse_iter = not self.start_from_left

        ops_left = list(circuit.operations_with_cycles(reverse=reverse_iter))
        print(
            f'Starting TreeScan with tree depth {self.tree_depth}'
            f' on circuit with {len(ops_left)} gates',
        )

        while ops_left:
            chunk = ops_left[:self.tree_depth]
            ops_left = ops_left[self.tree_depth:]

            all_circs = TreeScanningGateRemovalPass.get_tree_circs(
                circuit.num_cycles, circuit_copy, chunk,
            )

            _logger.debug(
                'Attempting removal of operation of up to'
                f' {self.tree_depth} operations.',
            )

            instantiated_circuits: list[Circuit] = await get_runtime().map(
                Circuit.instantiate,
                all_circs,
                target=target,
                **instantiate_options,
            )

            dists = [self.cost(c, target) for c in instantiated_circuits]

            # Pick least count with least dist
            for i, dist in enumerate(dists):
                if dist < self.success_threshold:
                    # Log gates removed
                    gate_dict_orig = circuit_copy.gate_counts
                    gate_dict_new = instantiated_circuits[i].gate_counts
                    gates_removed = {
                        k: circuit_copy.gate_counts[k] - gate_dict_new.get(k, 0)
                        for k in gate_dict_orig.keys()
                    }
                    gates_removed = {
                        k: v for k, v in gates_removed.items() if v != 0
                    }
                    _logger.debug(
                        f'Successfully removed {gates_removed} gates',
                    )
                    circuit_copy = instantiated_circuits[i]
                    break

        circuit.become(circuit_copy)


def default_collection_filter(op: Operation) -> bool:
    return True



"""This module implements the MachineModel class."""
from __future__ import annotations

from typing import Sequence
from typing import TYPE_CHECKING

from bqskit.compiler.gateset import GateSet
from bqskit.compiler.gateset import GateSetLike
from bqskit.ir.location import CircuitLocation
from bqskit.qis.graph import CouplingGraph
from bqskit.qis.graph import CouplingGraphLike
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_valid_radixes

if TYPE_CHECKING:
    from bqskit.ir.circuit import Circuit


class MachineModel:
    """A model of a quantum processing unit."""

    def __init__(
        self,
        num_qudits: int,
        coupling_graph: CouplingGraphLike | None = None,
        gate_set: GateSetLike | None = None,
        radixes: Sequence[int] = [],
    ) -> None:
        """
        MachineModel Constructor.

        Args:
            num_qudits (int): The total number of qudits in the machine.

            coupling_graph (Iterable[tuple[int, int]] | None): A coupling
                graph describing which pairs of qudits can interact.
                Given as an undirected edge set. If left as None, then
                an all-to-all coupling graph is used as a default.
                (Default: None)

            gate_set (GateSetLike | None): The native gate set available
                on the machine. If left as None, the default gate set
                will be used. See :func:`~GateSet.default_gate_set`.

            radixes (Sequence[int]): A sequence with its length equal
                to `num_qudits`. Each element specifies the base of a
                qudit. Defaults to qubits.

        Raises:
            ValueError: If `num_qudits` is nonpositive.

        Note:
            Pre-built models for many active QPUs exist in the
            :obj:`~bqskit.ext` package.
        """

        if not is_integer(num_qudits):
            raise TypeError(
                f'Expected integer num_qudits, got {type(num_qudits)}.',
            )

        if num_qudits <= 0:
            raise ValueError(f'Expected positive num_qudits, got {num_qudits}.')

        self.radixes = tuple(radixes if len(radixes) > 0 else [2] * num_qudits)

        if not is_valid_radixes(self.radixes):
            raise TypeError('Invalid qudit radixes.')

        if len(self.radixes) != num_qudits:
            raise ValueError(
                'Expected length of radixes to be equal to num_qudits:'
                ' %d != %d' % (len(self.radixes), num_qudits),
            )

        if coupling_graph is None:
            coupling_graph = CouplingGraph.all_to_all(num_qudits)

        if not CouplingGraph.is_valid_coupling_graph(
                coupling_graph, num_qudits,
        ):
            raise TypeError('Invalid coupling graph, expected list of tuples')

        if gate_set is None:
            gate_set = GateSet.default_gate_set(radixes)
        else:
            gate_set = GateSet(gate_set)

        if not isinstance(gate_set, GateSet):
            raise TypeError(f'Expected GateSet, got {type(gate_set)}.')

        self.gate_set = gate_set
        self.coupling_graph = CouplingGraph(coupling_graph)
        self.num_qudits = num_qudits

    def get_locations(self, block_size: int) -> list[CircuitLocation]:
        """Return all `block_size` connected blocks of qudit indicies."""
        return self.coupling_graph.get_subgraphs_of_size(block_size)

    def is_compatible(
        self,
        circuit: Circuit,
        placement: list[int] | None = None,
    ) -> bool:
        """Check if a circuit is compatible with this model."""
        if circuit.num_qudits > self.num_qudits:
            return False

        if any(g not in self.gate_set for g in circuit.gate_set):
            return False

        if placement is None:
            placement = list(range(circuit.num_qudits))

        if any(
            (placement[e[0]], placement[e[1]]) not in self.coupling_graph
            for e in circuit.coupling_graph
        ):
            return False

        if any(
            r != self.radixes[placement[i]]
            for i, r in enumerate(circuit.radixes)
        ):
            return False

        return True



"""This module implements the PauliZMatrices class."""
from __future__ import annotations

import itertools as it
from typing import Iterable
from typing import Iterator
from typing import overload
from typing import Sequence
from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt

from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_numeric
from bqskit.utils.typing import is_sequence

if TYPE_CHECKING:
    from bqskit.qis.unitary.unitary import RealVector


class PauliZMatrices(Sequence[npt.NDArray[np.complex128]]):
    """
    The group of Pauli Z matrices.

    A PauliZMatrices object represents the entire of set of diagonal Hermitian
    matrices for some number of qubits. These matrices are a linear combination
    of all n-fold tensor products of Pauli Z and the identity matrix.

    Examples:
    .. math::
        I + Z = \\begin{pmatrix}
            2 & 0 \\\\
            0 & 0 \\\\
        \\end{pmatrix}

    .. math::
        I \\otimes I + Z \\otimes I + 3 I \\otimes Z - Z \\otimes Z =
            \\begin{pmatrix}
                4 & 0 & 0 & 0 \\\\
                0 & -2 & 0 & 0 \\\\
                0 & 0 & 2 & 0 \\\\
                0 & 0 & 0 & -2 \\\\
            \\end{pmatrix}
    """

    Z = np.array(
        [
            [1, 0],
            [0, -1],
        ], dtype=np.complex128,
    )
    """The Pauli Z Matrix."""

    I = np.array(
        [
            [1, 0],
            [0, 1],
        ], dtype=np.complex128,
    )
    """The Identity Matrix."""

    def __init__(self, num_qudits: int) -> None:
        """
        Construct the Pauli Z group for `num_qudits` number of qubits.

        Args:
            num_qudits (int): Power of the tensor product of the Pauli Z
                group.

        Raises:
            ValueError: If `num_qudits` is less than or equal to 0.
        """

        if not is_integer(num_qudits):
            raise TypeError(
                'Expected integer for num_qudits, got %s.' %
                type(num_qudits),
            )

        if num_qudits <= 0:
            raise ValueError(
                'Expected positive integer for num_qudits, got %s.' % type(
                    num_qudits,
                ),
            )

        self.num_qudits = num_qudits

        if num_qudits == 1:
            self.paulizs = [
                PauliZMatrices.I,
                PauliZMatrices.Z,
            ]
        else:
            self.paulizs = []
            matrices = it.product(
                PauliZMatrices(
                    num_qudits - 1,
                ),
                PauliZMatrices(1),
            )
            for pauliz_n_1, pauliz_1 in matrices:
                self.paulizs.append(np.kron(pauliz_n_1, pauliz_1))  # type: ignore # noqa: E501

    def __iter__(self) -> Iterator[npt.NDArray[np.complex128]]:
        return self.paulizs.__iter__()

    @overload
    def __getitem__(self, index: int) -> npt.NDArray[np.complex128]:
        ...

    @overload
    def __getitem__(self, index: slice) -> list[npt.NDArray[np.complex128]]:
        ...

    def __getitem__(
        self,
        index: int | slice,
    ) -> npt.NDArray[np.complex128] | list[npt.NDArray[np.complex128]]:
        return self.paulizs[index]

    def __len__(self) -> int:
        return len(self.paulizs)

    @property
    def numpy(self) -> npt.NDArray[np.complex128]:
        """The NumPy array holding the pauliz matrices."""
        return np.array(self.paulizs)

    def __array__(
        self,
        dtype: np.typing.DTypeLike = np.complex128,
    ) -> npt.NDArray[np.complex128]:
        """Implements NumPy API for the PauliZMatrices class."""
        if dtype != np.complex128:
            raise ValueError('PauliZMatrices only supports Complex128 dtype.')

        return np.array(self.paulizs, dtype)

    def get_projection_matrices(
            self, q_set: Iterable[int],
    ) -> list[npt.NDArray[np.complex128]]:
        """
        Return the Pauli Z matrices that act only on qubits in `q_set`.

        Args:
            q_set (Iterable[int]): Active qubit indices

        Returns:
            list[np.ndarray]: Pauli Z matrices from `self` acting only
            on qubits in `q_set`.

        Raises:
            ValueError: if `q_set` is an invalid set of qubit indices.
        """
        q_set = list(q_set)

        if not all(is_integer(q) for q in q_set):
            raise TypeError('Expected sequence of integers for qubit indices.')

        if any(q < 0 or q >= self.num_qudits for q in q_set):
            raise ValueError('Qubit indices must be in [0, n).')

        if len(q_set) != len(set(q_set)):
            raise ValueError('Qubit indices cannot have duplicates.')

        # Nth Order Pauli Z Matrices can be thought of base 2 number
        # I = 0, Z = 1
        # IZZ = 1 * 2^2 + 1 * 2^1 + 0 * 2^0 = 6 (base 10)
        # This gives the idx of IZZ in paulizs
        # Note we read qubit index from the left,
        # so Z in ZII corresponds to q = 0
        pauliz_n_qubit = []
        for ps in it.product([0, 1], repeat=len(q_set)):
            idx = 0
            for p, q in zip(ps, q_set):
                idx += p * (2 ** (self.num_qudits - q - 1))
            pauliz_n_qubit.append(self.paulizs[idx])

        return pauliz_n_qubit

    def dot_product(self, alpha: RealVector) -> npt.NDArray[np.complex128]:
        """
        Computes the standard dot product of `alpha` with the paulis.

        Args:
            alpha (RealVector): The Pauli Z coefficients.

        Returns:
            np.ndarray: Sum of element-wise multiplication of `alpha`
            and `self.paulizs`.

        Raises:
            ValueError: If `alpha` and `self.paulizs` are incompatible.
        """

        if not is_sequence(alpha) or not all(is_numeric(a) for a in alpha):
            msg = f'Expected a sequence of numbers, got {type(alpha)}.'
            raise TypeError(msg)

        if len(alpha) != len(self):
            msg = (
                'Incorrect number of alpha values, expected '
                f'{len(self)}, got {len(alpha)}.'
            )
            raise ValueError(msg)

        return np.array(np.sum([a * s for a, s in zip(alpha, self.paulizs)], 0))

    @staticmethod
    def from_string(
        pauliz_string: str,
    ) -> npt.NDArray[np.complex128] | list[npt.NDArray[np.complex128]]:
        """
        Construct Pauli Z matrices from a string description.

        Args:
            pauliz_string (str): A string that describes the desired matrices.
                This is a comma-seperated list of Pauli Z strings.
                A Pauli Z string has the following regex pattern: [IZ]+

        Returns:
            np.ndarray | list[np.ndarray]: Either the single Pauli Z matrix
            if only one is constructed, or the list of the constructed
            Pauli Z matrices.

        Raises:
            ValueError: if `pauliz_string` is invalid.
        """

        if not isinstance(pauliz_string, str):
            msg = f'Expected str for pauliz_string, got {type(pauliz_string)}.'
            raise TypeError(msg)

        pauliz_strings = [
            string.strip().upper()
            for string in pauliz_string.split(',')
            if len(string.strip()) > 0
        ]

        pauliz_matrices = []
        idx_dict = {'I': 0, 'Z': 1}
        mat_dict = {
            'I': PauliZMatrices.I,
            'Z': PauliZMatrices.Z,
        }

        for pauli_string in pauliz_strings:
            if not all(char in 'IZ' for char in pauli_string):
                raise ValueError('Invalid Pauli Z string.')

            if len(pauli_string) <= 6:
                idx = 0
                for char in pauli_string:
                    idx *= 2
                    idx += idx_dict[char]
                pauliz_matrices.append(PauliZMatrices(len(pauli_string))[idx])
            else:
                acm = mat_dict[pauli_string[0]]
                for char in pauli_string[1:]:
                    acm = np.kron(acm, mat_dict[char])  # type: ignore
                pauliz_matrices.append(acm)

        if len(pauliz_matrices) == 1:
            return pauliz_matrices[0]

        return pauliz_matrices



"""This module implements the CompilationTask class."""
from __future__ import annotations

import uuid
from typing import TYPE_CHECKING

from bqskit.compiler.passdata import PassData
from bqskit.compiler.workflow import Workflow

if TYPE_CHECKING:
    from bqskit.ir.circuit import Circuit
    from bqskit.compiler.workflow import WorkflowLike


class CompilationTask():
    """
    A complete description of a quantum compilation task.

    The CompilationTask class describes a compilation workflow completely. These
    can be submitted to a BQSKit compiler to be efficiently executed.
    """

    def __init__(self, input: Circuit, workflow: WorkflowLike) -> None:
        """
        Construct a CompilationTask.

        Args:
            input (Circuit): The input circuit to be compiled.

            workflow (WorkflowLike): The configured workflow to be
                performed on the circuit.
        """
        self.task_id = uuid.uuid4()
        self.circuit = input
        self.workflow = Workflow(workflow)

        self.data = PassData(input)
        """The task's data for use in BQSKit passes."""

        self.done = False
        """True when the task is complete."""

        self.request_data = False
        """If true, :func:`run` will additionally return the PassData."""

        self.logging_level: int | None = None
        """A general filter on all logging messages in the system."""

        self.max_logging_depth = -1
        """No logging for tasks with more than `max_logging_depth` parents."""

    async def run(self) -> Circuit | tuple[Circuit, PassData]:
        """Execute the task."""
        await self.workflow.run(self.circuit, self.data)

        if not self.request_data:
            return self.circuit

        return self.circuit, self.data



"""This module implements helper methods for random generators."""
from __future__ import annotations

import ctypes
import random
from ctypes.util import find_library
from sys import platform

import numpy as np


def seed_random_sources(seed: int) -> None:
    """Seed the various sources of randomness BQSKit uses."""
    # set rand() seed, used by Ceres
    if platform != 'win32':
        libc = ctypes.CDLL(find_library('c'))
        libc.srand(seed)
    # set numpy seed
    np.random.seed(seed)
    random.seed(seed)



"""This module implements the GreedyPlacementPass class."""
from __future__ import annotations

import logging

import numpy as np

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
_logger = logging.getLogger(__name__)


class GreedyPlacementPass(BasePass):
    """Find a placement by starting with the most connected qudit."""

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""

        # Find physical qudit with highest degree
        graph = self.get_model(circuit, data).coupling_graph
        degrees = np.array(graph.get_qudit_degrees())
        highest_degree_qudit = int(np.argmax(degrees))

        # Search for placement by greedily adding best neighbor
        placement = [highest_degree_qudit]
        neighbors = graph.get_neighbors_of(highest_degree_qudit)
        while len(placement) < circuit.num_qudits:
            # Score each neighbor of current placement set
            best_score = None
            best_neighbor = None
            for q in neighbors:
                inter = [n for n in graph.get_neighbors_of(q) if n in placement]
                lookahead = sum(degrees[n] for n in graph.get_neighbors_of(q))
                score = (len(inter), degrees[q], lookahead)
                if best_score is None or score > best_score:
                    best_score = score
                    best_neighbor = q

            # Add best scoring neighbor to placement
            assert best_neighbor is not None
            neighbors.remove(best_neighbor)
            placement.append(best_neighbor)
            for n in graph.get_neighbors_of(best_neighbor):
                if n not in placement and n not in neighbors:
                    neighbors.append(n)

        data.placement = sorted(placement)

        _logger.info(f'Placed qudits on {data.placement}')

        # Raise an error if this is not a valid placement
        sg = graph.get_subgraph(data.placement)
        if not sg.is_fully_connected():
            raise RuntimeError('No valid placement found.')



"""This package implements BQSKit's post-processing passes."""
from __future__ import annotations

from bqskit.passes.processing.exhaustive import ExhaustiveGateRemovalPass
from bqskit.passes.processing.iterative import IterativeScanningGateRemovalPass
from bqskit.passes.processing.scan import ScanningGateRemovalPass
from bqskit.passes.processing.substitute import SubstitutePass
from bqskit.passes.processing.treescan import TreeScanningGateRemovalPass

__all__ = [
    'ExhaustiveGateRemovalPass',
    'IterativeScanningGateRemovalPass',
    'ScanningGateRemovalPass',
    'SubstitutePass',
    'TreeScanningGateRemovalPass',
]



"""This module implements the PassAlias class."""
from __future__ import annotations

import abc
from typing import TYPE_CHECKING

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.workflow import Workflow

if TYPE_CHECKING:
    from bqskit.compiler.passdata import PassData
    from bqskit.ir.circuit import Circuit


class PassAlias(BasePass):
    """A pass that is an alias for another pass or sequence of passes."""

    @abc.abstractmethod
    def get_passes(self) -> list[BasePass]:
        """Return the passes that should be run, when this alias is called."""

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        await Workflow(self.get_passes()).run(circuit, data)



"""This package contains class definitions for search based synthesis."""
from __future__ import annotations

from bqskit.passes.search.frontier import Frontier
from bqskit.passes.search.generator import LayerGenerator
from bqskit.passes.search.generators.simple import SimpleLayerGenerator
from bqskit.passes.search.heuristic import HeuristicFunction
from bqskit.passes.search.heuristics.astar import AStarHeuristic
from bqskit.passes.search.heuristics.dijkstra import DijkstraHeuristic
from bqskit.passes.search.heuristics.greedy import GreedyHeuristic

__all__ = [
    'SimpleLayerGenerator',
    'AStarHeuristic',
    'GreedyHeuristic',
    'DijkstraHeuristic',
    'Frontier',
    'LayerGenerator',
    'HeuristicFunction',
]



"""This module implments the CKMGate."""
from __future__ import annotations

import numpy as np
import numpy.typing as npt

from bqskit.ir.gates.qutritgate import QutritGate
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.cachedclass import CachedClass


class CKMGate(QutritGate, DifferentiableUnitary, CachedClass):
    """
    The Cabibbo-Kobayashi-Maskawa single qutrit gate.

    References:
        - Particle Data Group review:
          https://pdg.lbl.gov/2023/reviews/rpp2022-rev-ckm-matrix.pdf

        - https://link.springer.com/article/10.1007/JHEP01(2019)106

    Notes:
        - Due to the current experimental searches for the exact values
          of the mixing parameters, and to allow for beyond standard
          model physics, we provide the user with the flexibility of
          setting each parameter.
    """

    _num_qudits = 1
    _num_params = 4

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        self.check_parameters(params)
        # Parameters are mapped to the reference above as follows:
        #   - params[0] -> 1-3
        #   - params[1] -> 1-2
        #   - params[2] -> 2-3
        #   - params[3] -> cp

        s1 = np.sin(params[0])
        c1 = np.cos(params[0])
        s2 = np.sin(params[1])
        c2 = np.cos(params[1])
        s3 = np.sin(params[2])
        c3 = np.cos(params[2])

        p1 = np.exp(1j * params[3])
        m1 = np.exp(-1j * params[3])

        u1 = np.array([[1, 0, 0], [0, c3, s3], [0, -s3, c3]])
        u2 = np.array([[c1, 0, s1 * m1], [0, 1, 0], [-s1 * p1, 0, c1]])
        u3 = np.array([[c2, s2, 0], [-s2, c2, 0], [0, 0, 1]])

        return UnitaryMatrix(u1 @ u2 @ u3)

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        self.check_parameters(params)

        s1 = np.sin(params[0])
        c1 = np.cos(params[0])
        s2 = np.sin(params[1])
        c2 = np.cos(params[1])
        s3 = np.sin(params[2])
        c3 = np.cos(params[2])

        p1 = np.exp(1j * params[3])
        m1 = np.exp(-1j * params[3])

        u1 = np.array([[1, 0, 0], [0, c3, s3], [0, -s3, c3]])
        u2 = np.array([[c1, 0, s1 * m1], [0, 1, 0], [-s1 * p1, 0, c1]])
        u3 = np.array([[c2, s2, 0], [-s2, c2, 0], [0, 0, 1]])

        u1p = np.array([[0, 0, 0], [0, -s3, c3], [0, -c3, -s3]])
        u2p1 = np.array([
            [-s1, 0, c1 * m1], [0, 0, 0],
            [-c1 * m1, 0, s1],
        ])
        u2p2 = np.array([
            [0, 0, -1j * s1 * m1], [0, 0, 0],
            [-1j * s1 * p1, 0, 0],
        ])
        u3p = np.array([[s2, c2, 0], [-c2, -s2, 0], [0, 0, 0]])

        return np.array(
            [
                # wrt params[0] -> 1-3
                u1 @ u2p1 @ u3,

                # wrt params[1] -> 1-2
                u1 @ u2 @ u3p,

                # wrt params[2] -> 2-3
                u1p @ u2 @ u3,

                # wrt params[3] -> cp
                u1 @ u2p2 @ u3,
            ], dtype=np.complex128,
        )



"""This module implements the DaggerGate Class."""
from __future__ import annotations

import re

import numpy as np
import numpy.typing as npt

from bqskit.ir.gate import Gate
from bqskit.ir.gates.composed.daggergate import DaggerGate
from bqskit.ir.gates.composedgate import ComposedGate
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.docs import building_docs
from bqskit.utils.typing import is_integer


class PowerGate(
    ComposedGate,
    DifferentiableUnitary,
):
    """
    An arbitrary inverted gate.

    The PowerGate is a composed gate that equivalent to the
    integer power of the input gate.

    Examples:
        >>> from bqskit.ir.gates import TGate, TdgGate
        >>> PowerGate(TGate(),2).get_unitary() ==
           TdgGate().get_unitary()*TdgGate().get_unitary()
        True
    """

    def __init__(self, gate: Gate, power: int = 1) -> None:
        """
        Create a gate which is the integer power of the input gate.

        Args:
            gate (Gate): The Gate to conjugate transpose.
            power (int): The power index for the PowerGate.
        """
        if not isinstance(gate, Gate):
            raise TypeError('Expected gate object, got %s' % type(gate))

        if not is_integer(power):
            raise TypeError(f'Expected integer power, got {type(power)}.')

        self.gate = gate
        self.power = power
        self._name = f'[{gate.name}^{power}]'
        self._num_params = gate.num_params
        self._num_qudits = gate.num_qudits
        self._radixes = gate.radixes

        # If input is a constant gate, we can cache the unitary.
        if self.num_params == 0 and not building_docs():
            self.utry = self.gate.get_unitary([]).ipower(power)

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        if hasattr(self, 'utry'):
            return self.utry

        return self.gate.get_unitary(params).ipower(self.power)

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.

        Notes:
            The derivative of the integer power of matrix is equal
            to the derivative of the matrix multiplied by
            the integer-1 power of the matrix
            and by the integer power.
        """
        if hasattr(self, 'utry'):
            return np.array([])

        _, grad = self.get_unitary_and_grad(params)
        return grad

    def get_unitary_and_grad(
        self,
        params: RealVector = [],
    ) -> tuple[UnitaryMatrix, npt.NDArray[np.complex128]]:
        """
        Return the unitary and gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        # Constant gate case
        if hasattr(self, 'utry'):
            return self.utry, np.array([])

        grad_shape = (self.num_params, self.dim, self.dim)

        # Identity gate case
        if self.power == 0:
            utry = UnitaryMatrix.identity(self.dim)
            grad = np.zeros(grad_shape, dtype=np.complex128)
            return utry, grad

        # Invert the gate if the power is negative
        gate = self.gate if self.power > 0 else DaggerGate(self.gate)
        power = abs(self.power)

        # Parallel Dicts for unitary and gradient powers
        utrys = {}  # utrys[i] = gate^(2^i)
        grads = {}  # grads[i] = d(gate^(2^i))/d(params)

        # decompose the power as sum of powers of 2
        power_bin = bin(abs(power))[2:]
        binary_decomp = [
            len(power_bin) - 1 - xb.start()
            for xb in re.finditer('1', power_bin)
        ][::-1]
        max_power_of_2 = max(binary_decomp)

        # Base Case: 2^0
        utrys[0], grads[0] = gate.get_unitary_and_grad(params)  # type: ignore

        # Loop over powers of 2
        for i in range(1, max_power_of_2 + 1):
            # u^(2^i) = u^(2^(i-1)) @ u^(2^(i-1))
            utrys[i] = utrys[i - 1] @ utrys[i - 1]

            # d[u^(2^i)] = d[u^(2^(i-1)) @ u^(2^(i-1))] =
            grads[i] = grads[i - 1] @ utrys[i - 1] + utrys[i - 1] @ grads[i - 1]

        # Calculate binary composition of the unitary and gradient
        utry = utrys[binary_decomp[0]]
        grad = grads[binary_decomp[0]]
        for i in sorted(binary_decomp[1:]):
            grad = grad @ utrys[i] + utry @ grads[i]
            utry = utry @ utrys[i]

        return utry, grad

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, PowerGate)
            and self.gate == other.gate
            and self.power == other.power
        )

    def __hash__(self) -> int:
        return hash((self.power, self.gate))

    def get_inverse(self) -> Gate:
        """Return the gate's inverse as a gate."""
        return PowerGate(self.gate, -self.power)



"""This module implements the PauliZGate."""
from __future__ import annotations

from typing import Any

import numpy as np
import numpy.typing as npt

from bqskit.ir.gates.generalgate import GeneralGate
from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.pauliz import PauliZMatrices
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.docs import building_docs
from bqskit.utils.math import dexpmv
from bqskit.utils.math import dot_product
from bqskit.utils.math import pauliz_expansion
from bqskit.utils.math import unitary_log_no_i


class PauliZGate(QubitGate, DifferentiableUnitary, GeneralGate):
    """
    A gate representing an arbitrary diagonal rotation.

    This gate is given by:

    .. math::

        \\exp({i(\\vec{\\alpha} \\cdot \\vec{\\sigma_Z^{\\otimes n}})})

    Where :math:`\\vec{\\alpha}` are the gate's parameters,
    :math:`\\vec{\\sigma}` are the PauliZ Z matrices,
    and :math:`n` is the number of qubits this gate acts on.
    """

    def __init__(self, num_qudits: int) -> None:
        """
        Create a PauliZGate acting on `num_qudits` qubits.

        Args:
            num_qudits (int): The number of qudits this gate will act on.

        Raises:
            ValueError: If `num_qudits` is nonpositive.
        """

        if num_qudits <= 0:
            raise ValueError(f'Expected positive integer, got {num_qudits}')

        self._name = f'PauliZGate({num_qudits})'
        self._num_qudits = num_qudits
        paulizs = PauliZMatrices(self.num_qudits)
        self._num_params = len(paulizs)
        if building_docs():
            self.sigmav: npt.NDArray[Any] = np.array([])
        else:
            self.sigmav = (-1j / 2) * paulizs.numpy

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        self.check_parameters(params)
        H = dot_product(params, self.sigmav)
        eiH = np.diag(np.exp(np.diag(H)))
        return UnitaryMatrix(eiH, check_arguments=False)

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.

        TODO: Accelerated gradient computation for diagonal matrices.
        """
        self.check_parameters(params)
        H = dot_product(params, self.sigmav)
        _, dU = dexpmv(H, self.sigmav)
        return dU

    def get_unitary_and_grad(
        self,
        params: RealVector = [],
    ) -> tuple[UnitaryMatrix, npt.NDArray[np.complex128]]:
        """
        Return the unitary and gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        self.check_parameters(params)

        H = dot_product(params, self.sigmav)
        U, dU = dexpmv(H, self.sigmav)
        return UnitaryMatrix(U, check_arguments=False), dU

    def calc_params(self, utry: UnitaryMatrix) -> list[float]:
        """Return the parameters for this gate to implement `utry`"""
        return list(-2 * pauliz_expansion(unitary_log_no_i(utry.numpy)))

    def __eq__(self, o: object) -> bool:
        return isinstance(o, PauliZGate) and self.num_qudits == o.num_qudits

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.num_qudits))



"""This module implements BQSKit Runtime's Worker."""
from __future__ import annotations

import argparse
import logging
import os
import pickle
import signal
import sys
import time
import traceback
from dataclasses import dataclass
from multiprocessing import Process
from multiprocessing.connection import Client
from multiprocessing.connection import Connection
from queue import Empty
from queue import Queue
from threading import Lock
from threading import Thread
from typing import Any
from typing import Callable
from typing import cast
from typing import List
from typing import Sequence

from bqskit.runtime import default_worker_port
from bqskit.runtime import set_blas_thread_counts
from bqskit.runtime.address import RuntimeAddress
from bqskit.runtime.future import RuntimeFuture
from bqskit.runtime.message import RuntimeMessage
from bqskit.runtime.result import RuntimeResult
from bqskit.runtime.task import RuntimeTask


_logger = logging.getLogger(__name__)


@dataclass
class WorkerMailbox:
    """
    A mailbox on a worker is a final destination for a task's result.

    When a task is created, a mailbox is also created with an associated future.
    The parent task can await on the future, letting the worker's event loop
    know it is waiting on the associated result. When a result arrives, it is
    placed in the appropriate mailbox and the waiting task is placed into the
    ready queue.
    """
    expecting_single_result: bool = False
    expected_num_results: int = 0
    result: Any = None
    num_results: int = 0
    dest_addr: RuntimeAddress | None = None
    fresh_results: list[Any] | None = None

    @property
    def ready(self) -> bool:
        """Return true if the mailbox has all expected results."""
        return (
            self.num_results >= self.expected_num_results
            and self.num_results != 0
        )

    @property
    def has_task_waiting(self) -> bool:
        """Return True if a task is waiting on the result of this box."""
        return self.dest_addr is not None

    @staticmethod
    def new_mailbox(num_results: int | None = None) -> WorkerMailbox:
        """
        Create a new mailbox with `num_results` slots.

        If `num_results` is None (by default), then the mailbox will only have
        one slot and expect one result.
        """
        if num_results is None:
            return WorkerMailbox(True, 1)

        return WorkerMailbox(False, num_results, [None] * num_results)

    def get_new_results(self) -> list[tuple[int, Any]]:
        """Return and reset the results that have come in since previous
        call."""
        assert self.fresh_results is not None
        out = self.fresh_results
        self.fresh_results = []
        return out

    def deposit_result(self, result: RuntimeResult) -> None:
        """Store the result in the mailbox."""
        self.num_results += 1
        slot_id = result.return_address.mailbox_slot

        # Record as fresh result
        if self.fresh_results is None:
            self.fresh_results = []
        self.fresh_results.append((slot_id, result.result))

        if self.expecting_single_result:
            self.result = result.result
        else:
            self.result[slot_id] = result.result


class Worker:
    """
    BQSKit Runtime's Worker.

    BQSKit Runtime utilizes a dual-threaded worker to accept, execute,
    pause, spawn, resume, and complete tasks in a custom event loop built
    with python's async await mechanisms. Each worker receives and sends
    tasks and results to the greater system through a single duplex
    connection with a runtime server or manager. One thread performs
    work and sends outgoing messages, while the other thread handles
    incoming messages.

    At start-up, the worker receives an ID and waits for its first task.
    An executing task may use the `submit` and `map` methods to spawn child
    tasks and distribute them across the whole system. Once completed,
    those child tasks will have their results shipped back to the worker
    who created them. When a task awaits a child task, it is removed from
    the ready queue until the desired results come in.

    All created log records are shipped back to the client's process.
    This feature ensures compatibility with applications like jupyter
    that only print messages from the client process's stdout. Additionally,
    it allows BQSKit users seamless integration with the standard python
    logging module. From a user's perspective, they can configure any
    standard python logger from their process like usual and have the
    entire system honor that configuration. Lastly, we do support an
    additional logging option for maximum task depth. Tasks with more
    ancestors than the maximum logging depth will not produce any logs.

    Workers handle python errors by capturing and bubbling them up. For
    operating system-level crashes and errors -- such as seg-faults in
    client code -- the worker will attempt to print a stack trace and
    initiate a system-wide shutdown. However, note that these issues
    can go unhandled and cause the runtime to deadlock or crash.

    Workers perform very minimal scheduling of tasks. Newly created tasks
    are directly forwarded upwards to a manager or server, which in turn
    assigns them to workers. New tasks from above are commonly received
    in batches. In this case, all but one task from a batch is delayed.
    Delayed tasks are moved (in LIFO order) to the ready queue if a worker
    has no other work to complete. This mechanism encourages completing
    deeply-nested tasks first and prevents flooding the system with active
    tasks, which usually require much more memory than delayed ones.
    """

    def __init__(self, id: int, conn: Connection) -> None:
        """
        Initialize a worker with no tasks.

        Args:
            id (int): This worker's id.

            conn (Connection): This worker's duplex channel to a manager
                or a server.
        """
        self._id = id
        self._conn = conn

        self._tasks: dict[RuntimeAddress, RuntimeTask] = {}
        """Tracks all started, unfinished tasks on this worker."""

        self._delayed_tasks: list[RuntimeTask] = []
        """
        Store all delayed tasks in LIFO order.

        Delayed tasks have no context and are stored (more-or-less) as a
        function pointer together with the arguments. When it gets started, it
        consumes much more memory, so we delay the task start until necessary
        (at no cost)
        """

        self._ready_task_ids: Queue[RuntimeAddress] = Queue()
        """Tasks queued up for execution."""

        self._cancelled_task_ids: set[RuntimeAddress] = set()
        """To ensure newly-received cancelled tasks are never started."""

        self._active_task: RuntimeTask | None = None
        """The currently executing task if one is running."""

        self._running = True
        """Controls if the event loop is running."""

        self._mailboxes: dict[int, WorkerMailbox] = {}
        """Map from mailbox ids to worker mailboxes."""

        self._mailbox_counter = 0
        """This count ensures every mailbox has a unique id."""

        self._cache: dict[str, Any] = {}
        """Local worker cache."""

        self.most_recent_read_submit: RuntimeAddress | None = None
        """Tracks the most recently processed submit message from above."""

        self.read_receipt_mutex = Lock()
        """
        A lock to ensure waiting messages's read receipt is correct.

        This lock enforces atomic update of `most_recent_read_submit` and
        task addition/enqueueing. This is necessary to ensure that the
        idle status is always correct.
        """

        # Send out every client emitted log message upstream
        old_factory = logging.getLogRecordFactory()

        def record_factory(*args: Any, **kwargs: Any) -> logging.LogRecord:
            record = old_factory(*args, **kwargs)
            active_task = self._active_task
            if not record.name.startswith('bqskit.runtime'):
                if active_task is not None:
                    lvl = active_task.logging_level
                    if lvl is None or lvl <= record.levelno:
                        if lvl <= logging.DEBUG:
                            record.msg += f' [wid={self._id}'
                            items = active_task.log_context.items()
                            if len(items) > 0:
                                record.msg += ', '
                            con_str = ', '.join(f'{k}={v}' for k, v in items)
                            record.msg += con_str
                            record.msg += ']'
                        tid = active_task.comp_task_id
                        try:
                            serial = pickle.dumps(record)
                        except (pickle.PicklingError, TypeError):
                            serial = pickle.dumps((
                                record.name,
                                record.levelno,
                                record.getMessage(),
                            ))
                        self._conn.send((RuntimeMessage.LOG, (tid, serial)))
            return record

        logging.setLogRecordFactory(record_factory)

        # Start incoming thread
        self.incoming_thread = Thread(target=self.recv_incoming)
        self.incoming_thread.daemon = True
        self.incoming_thread.start()
        _logger.debug('Started incoming thread.')

        # Communicate that this worker is ready
        self._conn.send((RuntimeMessage.STARTED, self._id))

    def _loop(self) -> None:
        """Main worker event loop."""
        while self._running:
            try:
                self._try_step_next_ready_task()
            except Exception:
                self._running = False
                exc_info = sys.exc_info()
                error_str = ''.join(traceback.format_exception(*exc_info))
                _logger.error(error_str)
                try:
                    self._conn.send((RuntimeMessage.ERROR, error_str))
                except Exception:
                    pass

    def recv_incoming(self) -> None:
        """Continuously receive all incoming messages."""
        while self._running:
            # Receive message
            try:
                msg, payload = self._conn.recv()
            except Exception:
                _logger.debug('Crashed due to lost connection')
                if sys.platform == 'win32':
                    os.kill(os.getpid(), 9)
                else:
                    os.kill(os.getpid(), signal.SIGKILL)
                exit()

            _logger.debug(f'Received message {msg.name}.')
            _logger.log(1, f'Payload: {payload}')

            # Process message
            if msg == RuntimeMessage.SHUTDOWN:
                if sys.platform == 'win32':
                    os.kill(os.getpid(), 9)
                else:
                    os.kill(os.getpid(), signal.SIGKILL)

            elif msg == RuntimeMessage.SUBMIT:
                self.read_receipt_mutex.acquire()
                task = cast(RuntimeTask, payload)
                self.most_recent_read_submit = task.unique_id
                self._add_task(task)
                self.read_receipt_mutex.release()

            elif msg == RuntimeMessage.SUBMIT_BATCH:
                self.read_receipt_mutex.acquire()
                tasks = cast(List[RuntimeTask], payload)
                self.most_recent_read_submit = tasks[0].unique_id
                self._add_task(tasks.pop())  # Submit one task
                self._delayed_tasks.extend(tasks)  # Delay rest
                self.read_receipt_mutex.release()

            elif msg == RuntimeMessage.RESULT:
                result = cast(RuntimeResult, payload)
                self._handle_result(result)

            elif msg == RuntimeMessage.CANCEL:
                addr = cast(RuntimeAddress, payload)
                self._handle_cancel(addr)
                # TODO: preempt?

            elif msg == RuntimeMessage.COMMUNICATE:
                addrs, msg = cast(tuple[list[RuntimeAddress], Any], payload)
                self._handle_communicate(addrs, msg)

            elif msg == RuntimeMessage.IMPORTPATH:
                paths = cast(List[str], payload)
                for path in paths:
                    if path not in sys.path:
                        sys.path.append(path)

    def _add_task(self, task: RuntimeTask) -> None:
        """Start a task and add it to the loop."""
        self._tasks[task.return_address] = task
        task.start()
        self._ready_task_ids.put(task.return_address)

    def _handle_result(self, result: RuntimeResult) -> None:
        """Insert result into appropriate mailbox and wake waiting task."""
        assert result.return_address.worker_id == self._id

        mailbox_id = result.return_address.mailbox_index
        if mailbox_id not in self._mailboxes:
            # If the mailbox has been dropped due to a cancel, ignore result
            return

        box = self._mailboxes[mailbox_id]
        box.deposit_result(result)

        if box.has_task_waiting:
            assert box.dest_addr is not None
            task = self._tasks[box.dest_addr]

            if task.wake_on_next or box.ready:
                # print(f'Worker {self._id} is waking task
                # {task.return_address}, with {task.wake_on_next=},
                # {box.ready=}')
                self._ready_task_ids.put(box.dest_addr)  # Wake it
                box.dest_addr = None  # Prevent double wake

    def _handle_cancel(self, addr: RuntimeAddress) -> None:
        """
        Remove `addr` and its children tasks from this worker.

        Notes:
            Since `self._ready_task_ids' is a queue, it is more efficient
            to discard cancelled tasks when popping from it. Therefore, we
            do not do anything with `self._ready_task_ids` here.

            We also must make sure to call the `cancel` function of the
            tasks to make sure their coroutines are cleaned up.

            Also, we also don't need to send out cancel messages for
            cancelled children tasks since other workers can evaluate that
            for themselves using breadcrumbs and the original `addr` cancel
            message.
        """
        # TODO: Send update message?
        self._cancelled_task_ids.add(addr)

        # Remove all tasks that are children of `addr` from initialized tasks
        for key, task in list(self._tasks.items()):
            if task.is_descendant_of(addr):
                task.cancel()
                for mailbox_id in task.owned_mailboxes:
                    self._mailboxes.pop(mailbox_id)
                self._tasks.pop(key, None)

        # Remove all tasks that are children of `addr` from delayed tasks
        self._delayed_tasks = [
            t for t in self._delayed_tasks
            if not t.is_descendant_of(addr)
        ]

    def _handle_communicate(
        self,
        addrs: list[RuntimeAddress],
        msg: Any,
    ) -> None:
        for task_addr in addrs:
            if task_addr not in self._tasks:
                continue

            self._tasks[task_addr].msg_buffer.append(msg)

    def _get_next_ready_task(self) -> RuntimeTask | None:
        """Return the next ready task if one exists, otherwise block."""
        while True:
            if self._ready_task_ids.empty() and len(self._delayed_tasks) > 0:
                self._add_task(self._delayed_tasks.pop())
                continue

            # Critical section
            # Attempt to get a ready task. If none are available, message
            # the manager/server with a waiting message letting them
            # know the worker is idle. This needs to be atomic to prevent
            # the self.more_recent_read_submit from being updated after
            # catching the Empty exception, but before forming the payload.
            self.read_receipt_mutex.acquire()
            try:
                addr = self._ready_task_ids.get_nowait()

            except Empty:
                payload = (1, self.most_recent_read_submit)
                self._conn.send((RuntimeMessage.WAITING, payload))
                self.read_receipt_mutex.release()
                # Block for new message. Can release lock here since the
                # the `self.most_recent_read_submit` has been used.
                addr = self._ready_task_ids.get()

            else:
                self.read_receipt_mutex.release()

            # Handle a shutdown request that occured while waiting
            if not self._running:
                return None

            if addr in self._cancelled_task_ids or addr not in self._tasks:
                # When a task is cancelled on the worker it is not removed
                # from the ready queue because it is much cheaper to just
                # discard cancelled tasks as they come out.
                continue

            task = self._tasks[addr]

            if any(bcb in self._cancelled_task_ids for bcb in task.breadcrumbs):
                # If any of the selected tasks ancestor tasks are cancelled
                # then discard this one too. Each breadcrumb (bcb) is a
                # task address (unique system-wide task id) of an ancestor
                # task.
                # TODO: do I need to manually remove addr from self._tasks?
                continue

            return task

    def _try_step_next_ready_task(self) -> None:
        """Select a task to run, and advance it one step."""
        task = self._get_next_ready_task()

        if task is None:
            return

        try:
            self._active_task = task

            # Perform a step of the task and get the future it awaits on
            future = task.step(self._get_desired_result(task))

            self._process_await(task, future)

        except StopIteration as e:
            self._process_task_completion(task, e.value)

        except Exception as e:
            if type(e) is RuntimeError:
                for addr in self._cancelled_task_ids:
                    if task.is_descendant_of(addr):
                        return

            assert self._active_task is not None  # for type checker

            # Bubble up errors
            exc_info = sys.exc_info()
            error_str = ''.join(traceback.format_exception(*exc_info))
            error_payload = (self._active_task.comp_task_id, error_str)
            self._conn.send((RuntimeMessage.ERROR, error_payload))

        finally:
            self._active_task = None

    def _process_await(self, task: RuntimeTask, future: RuntimeFuture) -> None:
        """Process a task's await request."""
        if not isinstance(future, RuntimeFuture):
            raise RuntimeError('Can only await on a BQSKit RuntimeFuture.')

        if future.mailbox_id not in self._mailboxes:
            raise RuntimeError('Cannot await on a canceled task.')

        box = self._mailboxes[future.mailbox_id]

        # Let the mailbox know this task is waiting
        box.dest_addr = task.return_address
        task.desired_box_id = future.mailbox_id

        # if future._next_flag:
        #     # Set from Worker.next, implies the task wants the next result
        #     # if box.ready:
        #     #     m = 'Cannot wait for next results on a complete task.'
        #     #     raise RuntimeError(m)
        #     task.wake_on_next = True
        task.wake_on_next = future._next_flag
        # print(f'Worker {self._id} is waiting on task
        # {task.return_address}, with {task.wake_on_next=}')

        if box.ready:
            self._ready_task_ids.put(task.return_address)

    def _process_task_completion(self, task: RuntimeTask, result: Any) -> None:
        """Package and send out task result."""
        assert task is self._active_task
        packaged_result = RuntimeResult(task.return_address, result, self._id)

        if task.return_address not in self._tasks:
            # print(f'Task was cancelled: {task.return_address},
            # {task.fnargs[0].__name__}')
            return

        if task.return_address.worker_id == self._id:
            self._handle_result(packaged_result)
            self._conn.send((RuntimeMessage.UPDATE, -1))
            # Let manager know this worker has one less task
            # without sending a result
        else:
            self._conn.send((RuntimeMessage.RESULT, packaged_result))

        # Remove task
        self._tasks.pop(task.return_address, None)

        # Cancel any open tasks
        for mailbox_id in self._active_task.owned_mailboxes:
            # If task is complete, simply discard result
            if mailbox_id in self._mailboxes:
                if self._mailboxes[mailbox_id].ready:
                    self._mailboxes.pop(mailbox_id)
                    continue

            # Otherwise send a cancel message
            self.cancel(RuntimeFuture(mailbox_id))

    def _get_desired_result(self, task: RuntimeTask) -> Any:
        """Retrieve the task's desired result from the mailboxes."""
        if task.desired_box_id is None:
            return None

        box = self._mailboxes[task.desired_box_id]

        if task.wake_on_next:
            fresh_results = box.get_new_results()
            # assert len(fresh_results) > 0
            return fresh_results

        assert box.ready
        task.owned_mailboxes.remove(task.desired_box_id)
        return self._mailboxes.pop(task.desired_box_id).result

    def _get_new_mailbox_id(self) -> int:
        """Return a new unique mailbox id."""
        new_id = self._mailbox_counter
        self._mailbox_counter += 1
        return new_id

    def submit(
        self,
        fn: Callable[..., Any],
        *args: Any,
        task_name: str | None = None,
        log_context: dict[str, str] = {},
        **kwargs: Any,
    ) -> RuntimeFuture:
        """Submit `fn` as a task to the runtime."""
        assert self._active_task is not None

        if task_name is not None and not isinstance(task_name, str):
            raise RuntimeError('task_name must be a string.')

        if not isinstance(log_context, dict):
            raise RuntimeError('log_context must be a dictionary.')

        for k, v in log_context.items():
            if not isinstance(k, str) or not isinstance(v, str):
                raise RuntimeError(
                    'log_context must be a map from strings to strings.',
                )

        # Group fnargs together
        fnarg = (fn, args, kwargs)

        # Create a new mailbox
        mailbox_id = self._get_new_mailbox_id()
        self._mailboxes[mailbox_id] = WorkerMailbox.new_mailbox()
        self._active_task.owned_mailboxes.append(mailbox_id)

        # Create the task
        task = RuntimeTask(
            fnarg,
            RuntimeAddress(self._id, mailbox_id, 0),
            self._active_task.comp_task_id,
            self._active_task.breadcrumbs
            + (self._active_task.return_address,),
            self._active_task.logging_level,
            self._active_task.max_logging_depth,
            task_name,
            {**self._active_task.log_context, **log_context},
        )

        # Submit the task (on the next cycle)
        self._conn.send((RuntimeMessage.SUBMIT, task))

        # Return future pointing to the mailbox
        return RuntimeFuture(mailbox_id)

    def map(
        self,
        fn: Callable[..., Any],
        *args: Any,
        task_name: Sequence[str | None] | str | None = None,
        log_context: Sequence[dict[str, str]] | dict[str, str] = {},
        **kwargs: Any,
    ) -> RuntimeFuture:
        """Map `fn` over the input arguments distributed across the runtime."""
        assert self._active_task is not None

        if task_name is None or isinstance(task_name, str):
            task_name = [task_name] * len(args[0])

        if len(task_name) != len(args[0]):
            raise RuntimeError(
                'task_name must be a string or a list of strings equal'
                'in length to the number of tasks.',
            )

        if isinstance(log_context, dict):
            log_context = [log_context] * len(args[0])

        if len(log_context) != len(args[0]):
            raise RuntimeError(
                'log_context must be a dictionary or a list of dictionaries'
                ' equal in length to the number of tasks.',
            )

        for context in log_context:
            for k, v in context.items():
                if not isinstance(k, str) or not isinstance(v, str):
                    raise RuntimeError(
                        'log_context must be a map from strings to strings.',
                    )

        # Group fnargs together
        fnargs = []
        if len(args) == 1:
            for arg in args[0]:
                fnargs.append((fn, (arg,), kwargs))

        else:
            for subargs in zip(*args):
                fnargs.append((fn, subargs, kwargs))

        if len(fnargs) == 0:
            raise RuntimeError('Unable to map 0 tasks.')

        # Create a new mailbox
        mailbox_id = self._get_new_mailbox_id()
        self._mailboxes[mailbox_id] = WorkerMailbox.new_mailbox(len(fnargs))
        self._active_task.owned_mailboxes.append(mailbox_id)

        # Create the tasks
        breadcrumbs = self._active_task.breadcrumbs
        breadcrumbs += (self._active_task.return_address,)
        tasks = [
            RuntimeTask(
                fnarg,
                RuntimeAddress(self._id, mailbox_id, i),
                self._active_task.comp_task_id,
                breadcrumbs,
                self._active_task.logging_level,
                self._active_task.max_logging_depth,
                task_name[i],
                {**self._active_task.log_context, **log_context[i]},
            )
            for i, fnarg in enumerate(fnargs)
        ]

        # Submit the tasks
        self._conn.send((RuntimeMessage.SUBMIT_BATCH, tasks))

        # Return future pointing to the mailbox
        return RuntimeFuture(mailbox_id)

    def communicate(self, future: RuntimeFuture, msg: Any) -> None:
        """Send a message to the task associated with `future`."""
        assert self._active_task is not None
        assert future.mailbox_id in self._mailboxes

        num_slots = self._mailboxes[future.mailbox_id].expected_num_results
        addrs = [
            RuntimeAddress(self._id, future.mailbox_id, slot_id)
            for slot_id in range(num_slots)
        ]
        self._conn.send((RuntimeMessage.COMMUNICATE, (addrs, msg)))

    def get_messages(self) -> list[Any]:
        """Return all messages received by the worker for this task."""
        assert self._active_task is not None
        x = self._active_task.msg_buffer
        self._active_task.msg_buffer = []
        return x

    def cancel(self, future: RuntimeFuture) -> None:
        """Cancel all tasks associated with `future`."""
        assert self._active_task is not None
        num_slots = self._mailboxes[future.mailbox_id].expected_num_results
        self._active_task.owned_mailboxes.remove(future.mailbox_id)
        self._mailboxes.pop(future.mailbox_id)
        addrs = [
            RuntimeAddress(self._id, future.mailbox_id, slot_id)
            for slot_id in range(num_slots)
        ]
        for addr in addrs:
            self._conn.send((RuntimeMessage.CANCEL, addr))

    def get_cache(self) -> dict[str, Any]:
        """
        Retrieve worker's local cache.

        Returns:
            (dict[str, Any]): The worker's local cache. This cache can be
                used to store large or unserializable objects within a
                worker process' memory. Passes on the same worker that use
                the same object can load the object from this cache. If
                there are multiple workers, those workers will load their
                own copies of the object into their own cache.
        """
        return self._cache

    async def next(self, future: RuntimeFuture) -> list[tuple[int, Any]]:
        """
        Wait for and return the next batch of results from a map task.

        Returns:
            (list[tuple[int, Any]]): A list of the results that arrived
                since the last time this was called. On the first call,
                all results that have arrived since the task started are
                returned. Each result is paired with the index of its
                arguments in the original map call.
        """
        # if future._done:
        if future.mailbox_id not in self._mailboxes:
            raise RuntimeError('Cannot wait on an already completed result.')

        future._next_flag = True
        next_result_batch = await future
        future._next_flag = False
        return next_result_batch


# Global variable containing reference to this process's worker object.
_worker = None


def start_worker(
    w_id: int | None,
    port: int,
    cpu: int | None = None,
    logging_level: int = logging.WARNING,
    num_blas_threads: int = 1,
    log_client: bool = False,
) -> None:
    """Start this process's worker."""
    if w_id is not None:
        # Ignore interrupt signals on workers, boss will handle it for us
        # If w_id is None, then we are being spawned separately.
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        # TODO: check what needs to be done on win

    # Set number of BLAS threads
    set_blas_thread_counts(num_blas_threads)

    # Enforce no default logging
    logging.lastResort = logging.NullHandler()
    logging.getLogger().handlers.clear()

    # Pin worker to cpu
    if cpu is not None:
        if sys.platform == 'win32':
            raise RuntimeError('Cannot pin worker to cpu on windows.')
        os.sched_setaffinity(0, [cpu])

    # Connect to manager
    max_retries = 7
    wait_time = .1
    conn: Connection | None = None
    family = 'AF_INET' if sys.platform == 'win32' else None
    for _ in range(max_retries):
        try:
            conn = Client(('localhost', port), family)
        except (ConnectionRefusedError, TimeoutError):
            time.sleep(wait_time)
            wait_time *= 2
        else:
            break

    if conn is None:
        raise RuntimeError('Unable to establish connection with manager.')

    # If id isn't provided, wait for assignment
    if w_id is None:
        msg, w_id = conn.recv()
        assert isinstance(w_id, int)
        assert msg == RuntimeMessage.STARTED

    # Set up runtime logging
    if not log_client:
        _runtime_logger = logging.getLogger('bqskit.runtime')
    else:
        _runtime_logger = logging.getLogger()
    _runtime_logger.propagate = False
    _runtime_logger.setLevel(logging_level)
    _handler = logging.StreamHandler()
    _handler.setLevel(0)
    _fmt_header = '%(asctime)s.%(msecs)03d - %(levelname)-8s |'
    _fmt_message = f' [wid={w_id}]: %(message)s'
    _fmt = _fmt_header + _fmt_message
    _formatter = logging.Formatter(_fmt, '%H:%M:%S')
    _handler.setFormatter(_formatter)
    _runtime_logger.addHandler(_handler)

    # Build and start worker
    global _worker
    _worker = Worker(w_id, conn)
    _worker._loop()


def get_worker() -> Worker:
    """Return a handle on this process's worker."""
    if _worker is None:
        raise RuntimeError('Worker has not been started.')
    return _worker


def _check_positive(value: str) -> int:
    ivalue = int(value)
    if ivalue <= 0:
        raise argparse.ArgumentTypeError(
            '%s is an invalid positive int value' % value,
        )
    return ivalue


def start_worker_rank() -> None:
    """Entry point for spawning a rank of runtime worker processes."""
    parser = argparse.ArgumentParser(
        prog='bqskit-worker',
        description='Launch a rank of BQSKit runtime worker processes.',
    )
    parser.add_argument(
        'num_workers',
        type=_check_positive,
        help='The number of workers to spawn.',
    )
    parser.add_argument(
        '--cpus', '-c',
        nargs='+',
        type=int,
        help='Either one number or a list of numbers equal in length to the'
        ' number of workers. The workers will be pinned to specified logical'
        ' cpus. If a single-number is given, then all cpu indices are'
        ' enumerated starting at that number.',
    )
    parser.add_argument(
        '-p', '--port',
        type=int,
        default=default_worker_port,
        help='The port the workers will try to connect to a manager on.',
    )
    parser.add_argument(
        '-v', '--verbose',
        action='count',
        default=0,
        help='Enable logging of increasing verbosity, either -v, -vv, or -vvv.',
    )
    parser.add_argument(
        '-l', '--log-client',
        action='store_true',
        help='Log messages from the client process.',
    )
    parser.add_argument(
        '-t', '--num_blas_threads',
        type=int,
        default=1,
        help='The number of threads to use in BLAS libraries.',
    )
    args = parser.parse_args()

    if args.cpus is not None:
        if len(args.cpus) == 1:
            cpus = [args.cpus[0] + i for i in range(args.num_workers)]

        elif len(args.cpus) == args.num_workers:
            cpus = args.cpus

        else:
            raise RuntimeError(
                'The specified logical cpus are invalid. Expected either'
                ' a single number or a list of numbers equal in length to'
                ' the number of workers.',
            )

    else:
        cpus = [None for _ in range(args.num_workers)]

    logging_level = [30, 20, 10, 1][min(args.verbose, 3)]

    if args.log_client and logging_level > 10:
        raise RuntimeError('Cannot log client messages without at least -vv.')

    # Spawn worker process
    procs = []
    for cpu in cpus:
        pargs = (
            None,
            args.port,
            cpu,
            logging_level,
            args.num_blas_threads,
            args.log_client,
        )
        procs.append(Process(target=start_worker, args=pargs))
        procs[-1].start()

    # Join them
    for proc in procs:
        proc.join()



"""This module implements the UpdateDataPass class."""
from __future__ import annotations

import logging
from typing import Any

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit


_logger = logging.getLogger(__name__)


class UpdateDataPass(BasePass):
    """
    The UpdateDataPass class.

    The UpdateDataPass adds a key-value pair to data dictionary.
    """

    def __init__(self, key: str, val: Any) -> None:
        """
        Construct a UpdateDataPass.

        Args:
            key (str): The key to add.

            val (Any): The value to associate with the key.
        """

        if not isinstance(key, str):
            raise TypeError('Expected string for key, got %s.' % type(key))

        self.key = key
        self.val = val

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        _logger.debug(f'Injecting {self.key}:{self.val} into the data dict.')
        data[self.key] = self.val



"""This module contains functions to generate strategies from annotations."""
from __future__ import annotations

import collections
import inspect
import sys
from itertools import chain
from itertools import combinations
from typing import Any
from typing import Callable
from typing import Iterable
from typing import Sequence

import numpy as np
import pytest
from hypothesis import given
from hypothesis.extra.numpy import complex_number_dtypes
from hypothesis.extra.numpy import floating_dtypes
from hypothesis.extra.numpy import from_dtype
from hypothesis.strategies import booleans
from hypothesis.strategies import complex_numbers
from hypothesis.strategies import data
from hypothesis.strategies import dictionaries
from hypothesis.strategies import floats
from hypothesis.strategies import integers
from hypothesis.strategies import iterables
from hypothesis.strategies import just
from hypothesis.strategies import lists
from hypothesis.strategies import one_of
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import sets
from hypothesis.strategies import text
from hypothesis.strategies import tuples

from bqskit.utils.test.strategies import circuit_location_likes
from bqskit.utils.test.strategies import circuit_locations
from bqskit.utils.test.strategies import circuit_points
from bqskit.utils.test.strategies import circuit_regions
from bqskit.utils.test.strategies import circuits
from bqskit.utils.test.strategies import cycle_intervals
from bqskit.utils.test.strategies import everything_except
from bqskit.utils.test.strategies import gates
from bqskit.utils.test.strategies import operations
from bqskit.utils.test.strategies import unitaries
from bqskit.utils.test.strategies import unitary_likes


def _powerset(iterable: Iterable[Any]) -> Iterable[Any]:
    """
    Calculate the powerset of an iterable.

    Examples:

        >>> list(_powerset([1,2,3]))
        [(), (1,), (2,), (3,), (1, 2), (1, 3), (2, 3), (1, 2, 3)]

    References:

        https://stackoverflow.com/questions/18035595/powersets-in-python-using-
        itertools.
    """
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))


def _split_generic_arguments(args: str) -> list[str]:
    """Split a generic's type arguments up."""
    comma_indices = []
    num_open_brackets = 0
    for i, char in enumerate(args):
        if char == '[':
            num_open_brackets += 1
        elif char == ']':
            num_open_brackets -= 1
        elif char == ',' and num_open_brackets == 0:
            comma_indices.append(i)

    if len(comma_indices) == 0:
        return [args]

    to_return: list[str] = []
    last_index = 0
    for comma_index in comma_indices:
        to_return.append(args[last_index: comma_index])
        last_index = comma_index + 1
    to_return.append(args[last_index:])
    return to_return


def type_annotation_to_valid_strategy(annotation: str) -> SearchStrategy[Any]:
    """Convert a type annotation into a hypothesis strategy."""
    strategies: list[SearchStrategy[Any]] = []

    annotation = annotation.replace('RealVector', 'Sequence[float]')

    for type_str in annotation.split('|'):
        type_str = type_str.strip()

        if type_str == 'None':
            strategies.append(just(None))

        elif type_str == 'int':
            strategies.append(integers())

        elif type_str == 'float':
            strategies.append(floats())
            strategies.append(floating_dtypes().flatmap(from_dtype))

        elif type_str == 'complex':
            strategies.append(complex_numbers())
            strategies.append(complex_number_dtypes().flatmap(from_dtype))

        elif type_str == 'bool':
            strategies.append(booleans())

        elif type_str == 'str':
            strategies.append(text())

        elif type_str == 'Any':
            strategies.append(just(None))

        elif type_str.lower().startswith('tuple'):
            inner_strategies = []
            for arg in _split_generic_arguments(type_str[6:-1]):
                inner_strategies.append(type_annotation_to_valid_strategy(arg))
            strategies.append(tuples(*inner_strategies))

        elif type_str.lower().startswith('dict'):
            args = _split_generic_arguments(type_str[5:-1])
            key_strat = type_annotation_to_valid_strategy(args[0])
            val_strat = type_annotation_to_valid_strategy(args[1])
            strategies.append(dictionaries(key_strat, val_strat))

        elif type_str.lower().startswith('mapping'):
            args = _split_generic_arguments(type_str[8:-1])
            key_strat = type_annotation_to_valid_strategy(args[0])
            val_strat = type_annotation_to_valid_strategy(args[1])
            strategies.append(dictionaries(key_strat, val_strat))

        elif type_str.lower().startswith('list'):
            arg_strat = type_annotation_to_valid_strategy(type_str[5:-1])
            strategies.append(lists(arg_strat))

        elif type_str.lower().startswith('set'):
            arg_strat = type_annotation_to_valid_strategy(type_str[4:-1])
            strategies.append(sets(arg_strat))

        elif type_str.lower().startswith('sequence'):
            arg_strat = type_annotation_to_valid_strategy(type_str[9:-1])
            strategies.append(lists(arg_strat))

        elif type_str.lower().startswith('iterable'):
            arg_strat = type_annotation_to_valid_strategy(type_str[9:-1])
            strategies.append(iterables(arg_strat))

        elif type_str.lower().startswith('intervallike'):
            strat = type_annotation_to_valid_strategy('Tuple[int, int]')
            strategies.append(strat)
            strategies.append(cycle_intervals())

        elif type_str.lower().startswith('cycleinterval'):
            strategies.append(cycle_intervals())

        elif type_str.lower().startswith('circuitpointlike'):
            strat = type_annotation_to_valid_strategy('Tuple[int, int]')
            strategies.append(strat)
            strategies.append(circuit_points())

        elif type_str.lower().startswith('circuitpoint'):
            strategies.append(circuit_points())

        elif type_str.lower().startswith('circuitregionlike'):
            strat = type_annotation_to_valid_strategy('dict[int, IntervalLike]')
            strategies.append(strat)
            strategies.append(circuit_regions())

        elif type_str.lower().startswith('circuitregion'):
            strategies.append(circuit_regions())

        elif type_str.lower().startswith('unitarylike'):
            strategies.append(unitary_likes())

        elif type_str.lower().startswith('unitarymatrix'):
            strategies.append(unitaries())

        elif type_str.lower().startswith('gate'):
            strategies.append(gates())

        elif type_str.lower().startswith('operation'):
            strategies.append(operations())

        elif type_str.lower().startswith('circuitlocationlike'):
            strategies.append(circuit_locations())

        elif type_str.lower().startswith('circuitlocation'):
            strategies.append(circuit_location_likes())

        elif type_str.lower().startswith('circuit'):
            strategies.append(circuits(max_gates=1))

        else:
            raise ValueError(f'Cannot generate strategy for type: {type_str}')

    return one_of(strategies)


def type_annotation_to_invalid_strategy(annotation: str) -> SearchStrategy[Any]:
    """Convert a type annotation into an invalid hypothesis strategy."""
    strategies: list[SearchStrategy[Any]] = []
    types_to_avoid: set[type] = set()
    tuple_valids: dict[int, set[SearchStrategy[Any]]] = {}
    tuple_invalids: dict[int, set[SearchStrategy[Any]]] = {}
    dict_key_valids: set[SearchStrategy[Any]] = set()
    dict_key_invalids: set[SearchStrategy[Any]] = set()
    dict_val_valids: set[SearchStrategy[Any]] = set()
    dict_val_invalids: set[SearchStrategy[Any]] = set()
    list_invalids: set[SearchStrategy[Any]] = set()
    set_invalids: set[SearchStrategy[Any]] = set()
    iterable_invalids: set[SearchStrategy[Any]] = set()

    annotation = annotation.replace('RealVector', 'Sequence[float]')

    for type_str in annotation.split('|'):
        type_str = type_str.strip()

        if type_str == 'None':
            types_to_avoid.add(type(None))

        elif type_str == 'int':
            types_to_avoid.add(int)
            types_to_avoid.add(np.byte)
            types_to_avoid.add(np.short)
            types_to_avoid.add(np.intc)
            types_to_avoid.add(np.longlong)
            types_to_avoid.add(np.int8)
            types_to_avoid.add(np.int16)
            types_to_avoid.add(np.int32)
            types_to_avoid.add(np.int64)

        elif type_str == 'float':
            types_to_avoid.add(float)
            types_to_avoid.add(np.half)
            types_to_avoid.add(np.single)
            types_to_avoid.add(np.double)
            types_to_avoid.add(np.longdouble)
            types_to_avoid.add(np.float32)
            types_to_avoid.add(np.float64)

        elif type_str == 'complex':
            types_to_avoid.add(complex)
            types_to_avoid.add(np.csingle)
            types_to_avoid.add(np.cdouble)
            types_to_avoid.add(np.clongdouble)
            types_to_avoid.add(np.complex64)
            types_to_avoid.add(np.complex128)

        elif type_str == 'bool':
            types_to_avoid.add(bool)
            types_to_avoid.add(np.bool_)

        elif type_str == 'str':
            types_to_avoid.add(str)

        elif type_str == 'Any':
            continue

        elif type_str.lower().startswith('tuple'):
            args = _split_generic_arguments(type_str[6:-1])
            if len(args) not in tuple_valids:
                tuple_valids[len(args)] = set()
                tuple_invalids[len(args)] = set()

            for arg in args:
                valid_strat = type_annotation_to_valid_strategy(arg)
                invalid_strat = type_annotation_to_invalid_strategy(arg)
                tuple_valids[len(args)].add(valid_strat)
                tuple_invalids[len(args)].add(invalid_strat)

            types_to_avoid.add(tuple)

        elif type_str.lower().startswith('dict'):
            args = _split_generic_arguments(type_str[5:-1])
            dict_key_valids.add(type_annotation_to_valid_strategy(args[0]))
            dict_key_invalids.add(type_annotation_to_valid_strategy(args[1]))
            dict_val_valids.add(type_annotation_to_invalid_strategy(args[0]))
            dict_val_invalids.add(type_annotation_to_invalid_strategy(args[1]))
            types_to_avoid.add(dict)
            types_to_avoid.add(map)

        elif type_str.lower().startswith('mapping'):
            args = _split_generic_arguments(type_str[8:-1])
            dict_key_valids.add(type_annotation_to_valid_strategy(args[0]))
            dict_key_invalids.add(type_annotation_to_valid_strategy(args[1]))
            dict_val_valids.add(type_annotation_to_invalid_strategy(args[0]))
            dict_val_invalids.add(type_annotation_to_invalid_strategy(args[1]))
            types_to_avoid.add(dict)
            types_to_avoid.add(map)

        elif type_str.lower().startswith('list'):
            arg_strat = type_annotation_to_invalid_strategy(type_str[5:-1])
            list_invalids.add(arg_strat)
            types_to_avoid.add(list)

        elif type_str.lower().startswith('set'):
            arg_strat = type_annotation_to_invalid_strategy(type_str[4:-1])
            set_invalids.add(arg_strat)
            types_to_avoid.add(set)
            types_to_avoid.add(collections.abc.MutableSet)

        elif type_str.lower().startswith('sequence'):
            arg_strat = type_annotation_to_invalid_strategy(type_str[9:-1])
            list_invalids.add(arg_strat)
            types_to_avoid.add(Sequence)
            types_to_avoid.add(list)
            types_to_avoid.add(tuple)
            types_to_avoid.add(bytearray)
            types_to_avoid.add(bytes)

        elif type_str.lower().startswith('iterable'):
            arg_strat = type_annotation_to_invalid_strategy(type_str[9:-1])
            iterable_invalids.add(arg_strat)
            types_to_avoid.add(Sequence)
            types_to_avoid.add(list)
            types_to_avoid.add(tuple)
            types_to_avoid.add(Iterable)
            types_to_avoid.add(set)
            types_to_avoid.add(frozenset)
            types_to_avoid.add(dict)
            types_to_avoid.add(str)
            types_to_avoid.add(bytearray)
            types_to_avoid.add(bytes)
            types_to_avoid.add(collections.abc.MutableSet)
            types_to_avoid.add(enumerate)
            types_to_avoid.add(map)
            types_to_avoid.add(range)
            types_to_avoid.add(reversed)

        elif type_str.lower().startswith('intervallike'):
            types_to_avoid.add(tuple)

        elif type_str.lower().startswith('cycleinterval'):
            continue

        elif type_str.lower().startswith('circuitpointlike'):
            types_to_avoid.add(tuple)

        elif type_str.lower().startswith('circuitpoint'):
            continue

        elif type_str.lower().startswith('circuitregionlike'):
            types_to_avoid.add(dict)

        elif type_str.lower().startswith('circuitregion'):
            continue

        elif type_str.lower().startswith('circuitlocationlike'):
            types_to_avoid.add(int)
            types_to_avoid.add(Sequence)
            types_to_avoid.add(Iterable)
            types_to_avoid.add(list)
            types_to_avoid.add(tuple)
            types_to_avoid.add(collections.abc.MutableSet)
            types_to_avoid.add(enumerate)
            types_to_avoid.add(range)
            types_to_avoid.add(reversed)

        elif type_str.lower().startswith('circuitlocation'):
            continue

        elif type_str.lower().startswith('unitarylike'):
            types_to_avoid.add(np.ndarray)

        elif type_str.lower().startswith('unitarymatrix'):
            continue

        elif type_str.lower().startswith('gate'):
            continue

        elif type_str.lower().startswith('operation'):
            continue

        elif type_str.lower().startswith('circuit'):
            continue

        else:
            raise ValueError(f'Cannot generate strategy for type: {type_str}')

    strategies.append(everything_except(tuple(types_to_avoid)))

    for tuple_len in tuple_valids:
        for valid_set in _powerset(list(range(tuple_len))):  # (), (0,), (1,)
            if len(valid_set) == tuple_len:
                continue
            strategy_builder = []
            for i in range(tuple_len):
                if i in valid_set:
                    strat = one_of(list(tuple_valids[tuple_len]))
                    strategy_builder.append(strat)
                else:
                    strat = one_of(list(tuple_invalids[tuple_len]))
                    strategy_builder.append(strat)
            strategies.append(tuples(*strategy_builder))

    if len(dict_val_invalids) > 0:
        strategies.append(
            dictionaries(
                one_of(list(dict_key_valids)),
                one_of(list(dict_val_invalids)),
                min_size=1,
            ),
        )
        strategies.append(
            dictionaries(
                one_of(list(dict_key_invalids)),
                one_of(list(dict_val_valids)),
                min_size=1,
            ),
        )
        strategies.append(
            dictionaries(
                one_of(list(dict_key_invalids)),
                one_of(list(dict_val_invalids)),
                min_size=1,
            ),
        )

    if len(list_invalids) > 0:
        strategies.append(lists(one_of(list(list_invalids)), min_size=1))
    if len(set_invalids) > 0:
        strategies.append(sets(one_of(list(set_invalids)), min_size=1))
    if len(iterable_invalids) > 0:
        strategies.append(
            iterables(
                one_of(
                    list(iterable_invalids),
                ),
                min_size=1,
            ),
        )

    return one_of(strategies)


def invalid_type_test(
        func_to_test: Callable[..., Any],
        other_allowed_errors: list[type] = [],
) -> Callable[..., Callable[..., None]]:
    """
    Decorator to generate invalid type tests.

    A valid type test ensures that a function called with incorrect types
    does raise a TypeError.

    Examples:
        >>> class Foo:
        ...     def foo(self, x: int, y: int) -> None:
        ...         if not is_integer(x):
        ...             raise TypeError("")
        ...         if not is_integer(y):
        ...             raise TypeError("")

        >>> class TestFoo:
        ...     @invalid_type_test(Foo().foo)
        ...     def test_foo_invalid_type(self) -> None:
        ...         pass

        >>> @invalid_type_test(Foo().foo)
        ... def test_foo_invalid_type(self) -> None:
        ...     pass
    """
    if sys.version_info[0] == 3 and sys.version_info[1] < 9:
        return lambda x: x

    valids = []
    invalids = []
    for id, param in inspect.signature(func_to_test).parameters.items():
        if param.annotation == inspect._empty:
            raise ValueError(
                'Need type annotation to generate invalid type tests.',
            )

        valids.append(type_annotation_to_valid_strategy(param.annotation))
        invalids.append(type_annotation_to_invalid_strategy(param.annotation))

    strategies = []
    for valid_set in _powerset(list(range(len(valids)))):
        if len(valid_set) == len(valids):
            continue
        strategy_builder = []
        for i in range(len(valids)):
            if i in valid_set:
                strategy_builder.append(valids[i])
            else:
                strategy_builder.append(invalids[i])
        strategies.append(tuples(*strategy_builder))

    def inner(f: Callable[..., Any]) -> Callable[..., None]:
        if 'self' in inspect.signature(f).parameters:
            @pytest.mark.parametrize('strategy', strategies)
            @given(data=data())
            def invalid_type_test(self: Any, strategy: Any, data: Any) -> None:
                args = data.draw(strategy)
                with pytest.raises((TypeError,) + tuple(other_allowed_errors)):
                    func_to_test(*args)

            return invalid_type_test
        else:
            @pytest.mark.parametrize('strategy', strategies)
            @given(data=data())
            def invalid_type_test(strategy: Any, data: Any) -> None:
                args = data.draw(strategy)
                with pytest.raises((TypeError,) + tuple(other_allowed_errors)):
                    func_to_test(*args)

            return invalid_type_test

    return inner


def valid_type_test(
        func_to_test: Callable[..., Any],
) -> Callable[..., Callable[..., None]]:
    """
    Decorator to generate valid type tests.

    A valid type test ensures that a function called with correct types
    does not raise a TypeError.

    Examples:
        >>> class Foo:
        ...     def foo(self, x: int, y: int) -> None:
        ...         if not is_integer(x):
        ...             raise TypeError("")
        ...         if not is_integer(y):
        ...             raise TypeError("")

        >>> class TestFoo:
        ...     @valid_type_test(Foo().foo)
        ...     def test_foo_valid_type(self) -> None:
        ...         pass

        >>> @valid_type_test(Foo().foo)
        ... def test_foo_valid_type(self) -> None:
        ...     pass
    """
    if sys.version_info[0] == 3 and sys.version_info[1] < 9:
        return lambda x: x

    strategies = []
    for id, param in inspect.signature(func_to_test).parameters.items():
        if param.annotation == inspect._empty:
            raise ValueError(
                'Need type annotation to generate invalid type tests.',
            )

        strategies.append(type_annotation_to_valid_strategy(param.annotation))
    strategy = tuples(*strategies)

    def inner(f: Callable[..., Any]) -> Callable[..., None]:
        if 'self' in inspect.signature(f).parameters:
            @given(data=strategy)
            def valid_type_test(self: Any, data: Any) -> None:
                try:
                    func_to_test(*data)
                except TypeError:
                    assert False, 'Valid types caused TypeError.'
                except Exception:
                    pass

            return valid_type_test
        else:
            @given(data=strategy)
            def valid_type_test(data: Any) -> None:
                try:
                    func_to_test(*data)
                except TypeError:
                    assert False, 'Valid types caused TypeError.'
                except Exception:
                    pass

            return valid_type_test

    return inner



"""This package implements synthesis passes and synthesis related classes."""
from __future__ import annotations

from bqskit.passes.synthesis.bzxz import BlockZXZPass
from bqskit.passes.synthesis.bzxz import FullBlockZXZPass
from bqskit.passes.synthesis.diagonal import WalshDiagonalSynthesisPass
from bqskit.passes.synthesis.leap import LEAPSynthesisPass
from bqskit.passes.synthesis.pas import PermutationAwareSynthesisPass
from bqskit.passes.synthesis.qfast import QFASTDecompositionPass
from bqskit.passes.synthesis.qpredict import QPredictDecompositionPass
from bqskit.passes.synthesis.qsd import FullQSDPass
from bqskit.passes.synthesis.qsd import MGDPass
from bqskit.passes.synthesis.qsd import QSDPass
from bqskit.passes.synthesis.qsearch import QSearchSynthesisPass
from bqskit.passes.synthesis.synthesis import SynthesisPass
from bqskit.passes.synthesis.target import SetTargetPass

__all__ = [
    'LEAPSynthesisPass',
    'QFASTDecompositionPass',
    'QPredictDecompositionPass',
    'QSearchSynthesisPass',
    'SynthesisPass',
    'SetTargetPass',
    'PermutationAwareSynthesisPass',
    'WalshDiagonalSynthesisPass',
    'FullQSDPass',
    'MGDPass',
    'QSDPass',
    'BlockZXZPass',
    'FullBlockZXZPass',
]



"""This module implements the rotations due to the generatos of SU(3)."""
from __future__ import annotations

import numpy as np
import numpy.typing as npt

from bqskit.ir.gates.qutritgate import QutritGate
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.cachedclass import CachedClass
from bqskit.utils.typing import is_integer


class RSU3Gate(QutritGate, DifferentiableUnitary, CachedClass):
    """
    Rotation by SU3 generator for a single qutrit gate.

    .. math::
        \\exp(i * \\theta * \\lambda_j)

    where lambda_j is the j-th generator of the SU(3) Lie algebra.
    We use the physics notation, so each generator is Hermitian and not
    necessarily unitary. See See
    :func:`~bqskit.utils.math.compute_su_generators` for more info on the
    generators. There are N^2-1 = 3^2-1 = 8 generators and N-1 = 3-1 = 2
    mutually commute.

    References:
        - "Lie algebras in particle physics : from isospin to unified theories",
          Howard Georgi
    """

    _num_qudits = 1
    _num_params = 1

    def __init__(self, index: int) -> None:
        """
        Construct a RSU3Gate.

        Args:
            index (int): The index of the generator to use. See
                :func:`~bqskit.utils.math.compute_su_generators` for more
                info on the generators.

        Raises:
            ValueError: If index is less than 0 or greater than 7
        """
        if not is_integer(index):
            raise TypeError('RSU3Gate generator index must be an integer.')

        if index < 0 or index > 7:
            raise ValueError(f'Expected index between 0 and 7, got {index}.')

        self.index = index

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""

        self.check_parameters(params)

        cos = np.cos(params[0])
        sin = np.sin(params[0])

        matrix = np.eye(3, dtype=np.complex128)

        if self.index == 0:
            matrix[0, 0] = cos
            matrix[1, 1] = cos
            matrix[0, 1] = -1j * sin
            matrix[1, 0] = -1j * sin
        elif self.index == 1:
            matrix[0, 0] = cos
            matrix[1, 1] = cos
            matrix[0, 1] = -sin
            matrix[1, 0] = sin
        elif self.index == 2:
            matrix[0, 0] = np.exp(-1j * params[0])
            matrix[1, 1] = np.exp(1j * params[0])
        elif self.index == 3:
            matrix[0, 0] = cos
            matrix[2, 2] = cos
            matrix[0, 2] = -1j * sin
            matrix[2, 0] = -1j * sin
        elif self.index == 4:
            matrix[0, 0] = cos
            matrix[2, 2] = cos
            matrix[0, 2] = -sin
            matrix[2, 0] = sin
        elif self.index == 5:
            matrix[1, 1] = cos
            matrix[2, 2] = cos
            matrix[1, 2] = -1j * sin
            matrix[2, 1] = -1j * sin
        elif self.index == 6:
            matrix[1, 1] = cos
            matrix[2, 2] = cos
            matrix[1, 2] = -sin
            matrix[2, 1] = sin
        elif self.index == 7:
            matrix[0, 0] = np.exp(-1j * params[0] / np.sqrt(3))
            matrix[1, 1] = np.exp(-1j * params[0] / np.sqrt(3))
            matrix[2, 2] = np.exp(2j * params[0] / np.sqrt(3))
        return UnitaryMatrix(matrix, self.radixes)

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        self.check_parameters(params)

        cos = np.cos(params[0])
        sin = np.sin(params[0])

        matrix = np.zeros(
            (3, 3),
            dtype=np.complex128,
        )
        if self.index == 0:
            matrix[0, 0] = -sin
            matrix[1, 1] = -sin
            matrix[0, 1] = -1j * cos
            matrix[1, 0] = -1j * cos
        elif self.index == 1:
            matrix[0, 0] = -sin
            matrix[1, 1] = -sin
            matrix[0, 1] = -cos
            matrix[1, 0] = cos
        elif self.index == 2:
            matrix[0, 0] = -1j * np.exp(-1j * params[0])
            matrix[1, 1] = 1j * np.exp(1j * params[0])
        elif self.index == 3:
            matrix[0, 0] = -sin
            matrix[2, 2] = -sin
            matrix[0, 2] = -1j * cos
            matrix[2, 0] = -1j * cos
        elif self.index == 4:
            matrix[0, 0] = -sin
            matrix[2, 2] = -sin
            matrix[0, 2] = -cos
            matrix[2, 0] = cos
        elif self.index == 5:
            matrix[1, 1] = sin
            matrix[2, 2] = sin
            matrix[1, 2] = -1j * cos
            matrix[2, 1] = -1j * cos
        elif self.index == 6:
            matrix[1, 1] = sin
            matrix[2, 2] = sin
            matrix[1, 2] = -cos
            matrix[2, 1] = cos
        elif self.index == 7:
            matrix[0, 0] = -1j / np.sqrt(3) * \
                np.exp(-1j * params[0] / np.sqrt(3))
            matrix[1, 1] = -1j / np.sqrt(3) * \
                np.exp(-1j * params[0] / np.sqrt(3))
            matrix[2, 2] = 2j / np.sqrt(3) * np.exp(2j * params[0] / np.sqrt(3))
        return np.array([matrix], dtype=np.complex128)



"""
This module implements the Full Block ZXZ Decomposition.

Additionally, it defines the Block ZXZ Decomposition for a single pass.
"""
from __future__ import annotations

import logging
from typing import Any

import numpy as np
from scipy.linalg import eig
from scipy.linalg import svd

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.compiler.workflow import Workflow
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates import CircuitGate
from bqskit.ir.gates.constant import HGate
from bqskit.ir.gates.constant import IdentityGate
from bqskit.ir.gates.constant import ZGate
from bqskit.ir.gates.parameterized.mprz import MPRZGate
from bqskit.ir.location import CircuitLocation
from bqskit.ir.operation import Operation
from bqskit.passes.processing.extract_diagonal import ExtractDiagonalPass
from bqskit.passes.processing.scan import ScanningGateRemovalPass
from bqskit.passes.processing.treescan import TreeScanningGateRemovalPass
from bqskit.passes.synthesis.qsd import MGDPass
from bqskit.passes.synthesis.qsd import QSDPass
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.runtime import get_runtime


_logger = logging.getLogger(__name__)


class FullBlockZXZPass(BasePass):
    """
    A pass performing a full Block ZXZ decomposition.

    References:
        C.C. Paige, M. Wei,
        History and generality of the CS decomposition,
        Linear Algebra and its Applications,
        Volumes 208–209,
        1994,
        Pages 303-326,
        ISSN 0024-3795,
        https://doi.org/10.1016/0024-3795(94)90446-4.
    """

    def __init__(
        self,
        min_qudit_size: int = 2,
        perform_scan: bool = False,
        start_from_left: bool = True,
        tree_depth: int = 0,
        perform_extract: bool = True,
        instantiate_options: dict[str, Any] = {},
    ) -> None:
        """
        Perform the full Block ZXZ decomposition.

        Args:
            min_qudit_size (int): Performs QSD until the circuit only has
                VariableUnitaryGates with a number of qudits less than or equal
                to this value. (Default: 2)
            perform_scan (bool): Whether or not to perform the scanning
                gate removal pass. (Default: False)
            start_from_left (bool): Determines where the scan starts
                attempting to remove gates from. If True, scan goes left
                to right, otherwise right to left. (Default: True)
            tree_depth (int): The depth of the tree to use in the
                TreeScanningGateRemovalPass. If set to 0, we will instead
                use the ScanningGateRemovalPass. (Default: 0)
            perform_extract (bool): Whether or not to perform the diagonal
                extraction pass. (Default: True)
            instantiate_options (dict): The options to pass to the
                scanning gate removal pass. (Default: {})
        """

        self.start_from_left = start_from_left
        self.min_qudit_size = min_qudit_size
        instantiation_options = {'method': 'qfactor'}
        instantiation_options.update(instantiate_options)
        self.perform_scan = perform_scan
        self.perform_extract = perform_extract
        self.scan = ScanningGateRemovalPass(
            start_from_left=start_from_left,
            instantiate_options=instantiation_options,
        )
        if tree_depth > 0:
            self.scan = TreeScanningGateRemovalPass(
                start_from_left=start_from_left,
                instantiate_options=instantiation_options,
                tree_depth=tree_depth,
            )
        self.bzxz = BlockZXZPass(min_qudit_size=min_qudit_size)
        self.mgd = MGDPass()
        self.diag = ExtractDiagonalPass(qudit_size=min_qudit_size)

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """
        Perform succesive rounds of Block ZXZ decomposition until the circuit is
        fully decomposed with no VariableUnitaryGates larger then
        `min_qudit_size`.

        At the end, attempt to extrac the diagonal gate and commute through the
        circuit to find optimal CNOT counts.
        """
        passes: list[BasePass] = []
        start_num = max(x.num_qudits for x in circuit.operations())

        for _ in range(self.min_qudit_size, start_num):
            passes.append(self.bzxz)
            if self.perform_scan:
                passes.append(self.scan)

        if self.perform_extract:
            passes.append(self.diag)

        # Once we have commuted the diagonal gate, we can break down the
        # multiplexed gates
        for _ in range(1, start_num):
            passes.append(self.mgd)
            if self.perform_scan:
                passes.append(self.scan)

        await Workflow(passes).run(circuit, data)


class BlockZXZPass(BasePass):
    """
    A pass performing one round of decomposition from the Block ZXZ algorithm.

    References:
        Krol, Anna M., and Zaid Al-Ars.
        "Highly Efficient Decomposition of n-Qubit Quantum Gates
        Based on Block-ZXZ Decomposition."
        arXiv preprint arXiv:2403.13692 (2024).
        https://arxiv.org/html/2403.13692v1
    """

    def __init__(
        self,
        min_qudit_size: int = 4,
    ) -> None:
        """
        Construct a single level of the QSDPass.

        Args:
            min_qudit_size (int): Performs a decomposition on all gates
                with width > min_qudit_size
        """
        self.min_qudit_size = min_qudit_size

    @staticmethod
    def initial_decompose(U: UnitaryMatrix) -> tuple[
        UnitaryMatrix,
        UnitaryMatrix,
        UnitaryMatrix,
        UnitaryMatrix,
    ]:
        """
        This function decomposes the a given unitary U into 2 controlled
        unitaries (B, and C), 1 multiplexed unitary (A) and 2 Hadamard gates.

        We return the A1 and A2 (sub-unitaries of the multipelexed unitary A),
        as well as the sub unitary B and C.

        By sub-unitary, we mean the unitary that will be applied to the bottom
        n - 1 qubits given the value of the top qubit 0. For a controlled
        unitary, that sub-unitary is either the Identity (if qubit 0 is 0) or a
        unitary on n-1 qubits if qubit 0 is 1.

        This follows equations 5-9 in the paper.
        """
        # To solve for A1, A2, and C, we must divide U into 4 submatrices.
        # The upper left block is X, the upper right block is Y, the lower left
        # block is U_21, and the lower right block is U_22.

        X = U[0:len(U) // 2, 0:len(U) // 2]
        Y = U[0:len(U) // 2, len(U) // 2:]
        U_21 = U[len(U) // 2:, 0:len(U) // 2]
        U_22 = U[len(U) // 2:, len(U) // 2:]

        # We let X = V_X @ s_X @ W_X where V_X and W_X are unitary and s_X is
        # a diagonal matrix of singular values. We use the SVD to find these
        # matrices. We do the same decomposition for Y.
        V_X, s_X, W_XH = svd(X)
        V_Y, s_Y, W_YH = svd(Y)

        # We can then define S_X and S_Y as V_X @ s_X @ V_X^H and
        # V_Y @ s_Y @ V_Y^H
        S_X = V_X @ np.diag(s_X) @ V_X.conj().T
        S_Y = V_Y @ np.diag(s_Y) @ V_Y.conj().T

        # We define U_X and U_Y as V_X^H @ U @ W_X and V_Y^H @ U @ W_Y
        U_X = V_X @ W_XH
        U_Y = V_Y @ W_YH

        # Now, to perform the decomposition as defined in Section 4.1
        # We can set C dagger as i(U_Y^H @ U_X)
        CH = 1j * U_Y.conj().T @ U_X

        # We then set A_1 as S_X @ U_X and A_2 as U_21 + U_22 @ (iU_Y^H @ U_X)
        A_1 = UnitaryMatrix((S_X + 1j * S_Y) @ U_X)
        A_2 = UnitaryMatrix(U_21 + U_22 @ (1j * U_Y.conj().T @ U_X))

        # Finally, we can set B as 2(A_1^H @ U) - I
        I = np.eye(len(U) // 2)
        B = UnitaryMatrix((2 * (A_1.conj().T @ X)) - I)

        return A_1, A_2, B, UnitaryMatrix(CH.conj().T)

    @staticmethod
    def demultiplex(
        U_1: UnitaryMatrix,
        U_2: UnitaryMatrix,
    ) -> tuple[
        UnitaryMatrix,
        RealVector,
        UnitaryMatrix,
    ]:
        """
        Demultiplex U that is block_diag(U_1, U_2) to 2 Unitary Matrices V and W
        and a diagonal matrix D.

        The Diagonal Matrix D corresponds to a Multiplexed Rz gate acting on the
        most significant qubit.

        We return the Unitary Matrices V and W and the parameters for the
        corresponding MPRZ gate.
        """
        # U can be decomposed into U = (I otimes V )(D otimes D†)(I otimes W )

        # We can find V,D^2 by performing an eigen decomposition of
        # U_1 @ U_2†
        d2, V = eig(U_1 @ U_2.conj().T)
        d = np.sqrt(d2)
        D = np.diag(d)

        # We can then multiply to solve for W
        W = D @ V.conj().T @ U_2

        # We can use d to find the parameters for the MPRZ gate.
        # Note than because and Rz does exp(-i * theta / 2), we must
        # multiply by -2.
        d_params: list[float] = list(np.angle(d) * -2)

        return UnitaryMatrix(V), d_params, UnitaryMatrix(W)

    @staticmethod
    def zxz(orig_u: UnitaryMatrix) -> Circuit:
        """Return the circuit that is generated from one levl of Block ZXZ
        decomposition."""

        # First calculate the A, B, and C matrices for the initial decomp
        A_1, A_2, B, C = BlockZXZPass.initial_decompose(orig_u)

        # Now decompose thee multiplexed A gate and the controlled C gate
        I = IdentityGate(orig_u.num_qudits - 1).get_unitary()
        VA, AZ_params, WA = BlockZXZPass.demultiplex(A_1, A_2)
        VC, CZ_params, WC = BlockZXZPass.demultiplex(I, C)

        # Now calculate optimized B_tilde (Section 5.2) and decompose
        # We merge in WA and VC into B and then merge in the additional
        # CZ gates from the optimization
        small_I = IdentityGate(orig_u.num_qudits - 2).get_unitary()
        Z = ZGate(1).get_unitary()
        B_tilde_1 = UnitaryMatrix(WA @ VC)
        B_tilde_2 = UnitaryMatrix(
            np.kron(Z, small_I) @
            WA @ B @ VC @ np.kron(Z, small_I),
        )
        VB, BZ_params, WB = BlockZXZPass.demultiplex(B_tilde_1, B_tilde_2)

        # Define circuit locations
        # We let the target qubit be qubit 0 (top qubit in diagram)
        # The select qubits are the remaining qubits
        controlled_qubit = 0
        select_qubits = list(range(1, orig_u.num_qudits))
        all_qubits = list(range(0, orig_u.num_qudits))

        # Construct Circuit
        circ = Circuit(orig_u.num_qudits)

        # Add WC gate
        wc_gate, wc_params = QSDPass.create_unitary_gate(WC)
        circ.append_gate(wc_gate, CircuitLocation(select_qubits), wc_params)

        # Add decomposed MPRZ gate circuit.
        # Since the MPRZ gate circuit sets the target qubit as qubit
        # num_qudits - 1, we must shift the qubits to the left
        shifted_qubits = all_qubits[1:] + all_qubits[0:1]
        circ.append_circuit(
            MGDPass.decompose_mpx_two_levels(
                decompose_ry=False,
                params=CZ_params,
                num_qudits=orig_u.num_qudits,
                drop_last_cnot=True,
            ),
            shifted_qubits,
        )

        # Add the decomposed B-tilde gates WB and a Hadamard
        combo_1_gate, combo_1_params = QSDPass.create_unitary_gate(WB)
        circ.append_gate(
            combo_1_gate, CircuitLocation(select_qubits),
            combo_1_params,
        )
        circ.append_gate(HGate(), CircuitLocation((controlled_qubit,)))

        # The central MPRZ_gate. We set the target to the controlled qubit,
        # so there is no need to shift
        z_gate = MPRZGate(len(all_qubits), 0)
        circ.append_gate(z_gate, CircuitLocation(all_qubits), BZ_params)

        # Now add the decomposed B-tilde gates VB and a Hadamard
        combo_2_gate, combo_2_params = QSDPass.create_unitary_gate(VB)
        circ.append_gate(
            combo_2_gate, CircuitLocation(select_qubits),
            combo_2_params,
        )
        circ.append_gate(HGate(), CircuitLocation((controlled_qubit,)))

        # Add the decomposed MPRZ gate circuit again on shifted qubits
        circ.append_circuit(
            MGDPass.decompose_mpx_two_levels(
                decompose_ry=False,
                params=AZ_params,
                num_qudits=orig_u.num_qudits,
                reverse=True,
                drop_last_cnot=True,
            ),
            shifted_qubits,
        )

        va_gate, va_params = QSDPass.create_unitary_gate(VA)
        circ.append_gate(va_gate, CircuitLocation(select_qubits), va_params)

        # assert np.allclose(orig_u, circ.get_unitary())
        return circ

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform a single level of the Block ZXZ decomposition."""
        unitaries, pts, locations = QSDPass.get_variable_unitary_pts(
            circuit, self.min_qudit_size,
        )

        if len(unitaries) > 0:
            circs = await get_runtime().map(BlockZXZPass.zxz, unitaries)
            # Do bulk replace
            circ_gates = [CircuitGate(x) for x in circs]
            circ_ops = [
                Operation(x, locations[i], x._circuit.params)
                for i, x in enumerate(circ_gates)
            ]
            circuit.batch_replace(pts, circ_ops)
            circuit.unfold_all()

        circuit.unfold_all()
